import birdepy.squares_expm as _squares_expm
import birdepy.squares_fm as _squares_fm
import birdepy.squares_gwa as _squares_gwa
import birdepy.utility as ut


def discrete_est_lse(data, squares, model, b_rate, d_rate, z_trunc,
                     idx_known_p, known_p, p0, p_bounds, con, opt_method,
                     options):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using least squares estimation.
    See :ref:`here <Least Squares Estimation>` for more information.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'lse'::

        bd.estimate(t_data, p_data, p0, p_bounds, framework='lse', squares='fm', z_trunc=())

    The parameters associated with this framework (listed below) can be
    accessed using kwargs in :func:`birdepy.estimate()`. See documentation of
    :func:`birdepy.estimate` for the main arguments.

    Parameters
    ----------
    squares : string, optional
        Squared error from the expected value approximation method. Should be
        one of (alphabetical ordering):

            - 'expm'
            - 'fm' (default)
            - 'gwa'

    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default
        ``z_trunc=[z_min, z_max]`` where
        ``z_min=max(0, min(p_data) - 2*z_range)``
        and ``z_max=max(p_data) + 2*z_range`` where
        ``z_range=max(p_data)-min(p_data)``. Only applicable to `squares`
        method 'expm'.

    Examples
    --------
    Simulate a sample path and estimate the parameters using the various
    likelihood approximation methods: ::

        import birdepy as bd
        t_data = list(range(100))
        p_data = bd.simulate.discrete([0.75, 0.25, 0.02, 1], 'Ricker', 10, t_data,
                                      survival=True, seed=2021)
        for squares in ['expm', 'fm', 'gwa']:
            est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[1e-6,1], [1e-6,1], [1e-6, 0.1]],
                              framework='lse', model='Ricker', idx_known_p=[3], known_p=[1],
                              squares=squares)
            print('lse estimate using', squares, 'is', est.p, 'computed in ', est.compute_time, 'seconds.')

    Outputs: ::

        lse estimate using expm is [0.7879591925854611, 0.26289368236374644, 0.02000871498805996] computed in  1.2579967975616455 seconds.
        lse estimate using fm is [0.7941603523732229, 0.2766621569715867, 0.019363240909074483] computed in  5.483000755310059 seconds.
        lse estimate using gwa is [0.7024952317382023, 0.20563650045779058, 0.022598851311981704] computed in  0.09800028800964355 seconds.

    The constraint ``con={'type': 'ineq', 'fun': lambda p: p[0]-p[1]}`` ensures
    that p[0] > p[1] (i.e., rate of spread greater than recovery rate).

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """

    # Define a function that computes the difference between the observations and the approximate
    # expected value of the process
    if squares == 'fm':
        sq = _squares_fm.sq_bld(data, b_rate, d_rate)
    elif squares == 'expm':
        sq = _squares_expm.sq_bld(data, b_rate, d_rate, z_trunc)
    elif squares == 'gwa':
        sq = _squares_gwa.sq_bld(data, b_rate, d_rate)
    else:
        raise TypeError("Argument squares has an unknown value.")

    # Adjust the error function to account for unknown parameters
    def error_fun(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return sq(param)

    return ut.minimize_(error_fun, p0, p_bounds, con, opt_method, options)



import numpy as np
import birdepy.utility as ut


def probability_Erlang(z0, zt, t, param, b_rate, d_rate, z_trunc, k):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *Erlangization* method.

    To use this function call ``birdepy.probability`` with `method` set to
    'Erlang'::

        bd.probability(z0, zt, t, param, method='Erlang', z_trunc=(), k=1502)

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------
    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default
        ``z_trunc=[z_min, z_max]`` where ``z_min=max(0, min(z0, zt) - 100)``
        and ``z_max=max(z0, zt) + 100``

    k : int, optional
        Number of terms to include in approximation to probability.

    Examples
    --------
    Approximate transition probability for a Verhulst model using Erlangization: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='Erlang')[0][0]

    Outputs: ::

        0.002731464736623327

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    For a method related to this one see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`


    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Asmussen, S., Avram, F. and Usabel, M. Erlangian approximations for
     finite-horizon ruin probabilities. ASTIN Bulletin: The Journal of the IAA,
     32(2):267-281, 2002.

    """

    # Extract the min and max population sizes considered
    z_min, z_max = z_trunc

    # Count how many states are considered
    num_states = int(z_max - z_min + 1)

    # If more than one time is requested it is easiest to divert into a different code block
    if t.size == 1:
        # Determine the transition rate matrix
        q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)
        # Determine the associated 'R' matrix of discrete time transitions at Erlang distributed
        # epochs
        r_mat = (k / t) * np.linalg.inv((k / t) * np.eye(num_states) - q_mat)
        # Take k Erlang distributed epochs so the considered time converges to the requested t
        # with low variance
        p_mat = np.linalg.matrix_power(r_mat, k)
        # Fill in the output according to requested initial and final states (although many more
        # probabilities are actually computed)
        output = p_mat[np.ix_(np.array(z0 - z_min, dtype=np.int32),
                              np.array(zt - z_min, dtype=np.int32))]
    else:
        # Initialize an output array to be filled in as we loop over times
        output = np.zeros((t.size, z0.size, zt.size))
        # Determine the transition rate matrix
        q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)
        for idx in range(t.size):
            # Determine the associated 'R' matrix of discrete time transitions at Erlang distributed
            # epochs
            r_mat = (k / t[idx]) * np.linalg.inv((k / t[idx]) * np.eye(num_states) - q_mat)
            # Take k Erlang distributed epochs so the considered time converges to the requested t
            # with low variance
            p_mat = np.linalg.matrix_power(r_mat, k)
            # Fill in the output according to requested initial and final states (although many more
            # probabilities are actually computed)
            output[idx, :, :] = p_mat[np.ix_(np.array(z0 - z_min, dtype=np.int32),
                                             np.array(zt - z_min, dtype=np.int32))]
    return output



import numpy as np
from scipy import integrate as int
from scipy.stats import norm
import warnings
import birdepy.utility as ut


def probability_da(z0, zt, t, param, b_rate, d_rate, h_fun, k):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *diffusion approximation* method.

    To use this function call :func:`birdepy.probability` with `method` set to
    'da'::

        bd.probability(z0, zt, t, param, method='da', k=10)

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------
    k : int, optional
        ``2**k + 1`` equally spaced samples between observation times are used
        to compute integrals.

    Examples
    --------
    Approximate transition probability for a Verhulst model using diffusion approximation: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.01, 0.01], model='Verhulst', method='da')[0][0])

    Outputs: ::

        0.002266101391343583

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    For a method related to this one see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Ross, J.V., Pagendam, D.E. and Pollett, P.K. On parameter estimation
     in population models II: Multi-dimensional processes and transient
     dynamics. Theoretical Population Biology, 75(2-3):123-132, 2009.

    """
    # Determine a partition for numerical integration
    t_linspace = np.union1d(np.linspace(0, t[-1], t.size * (2 ** k + 1)), t)

    # These are methods used to solve differential equations as used by scipy.integrate.solve_ivp()
    solver_methods = ['RK45', 'Radau', 'RK23', 'BDF', 'DOP853']

    # Output format depends on whether multiple times are requested
    if t.size == 1:
        output = np.zeros((z0.size, zt.size))
    else:
        output = np.zeros((t.size, z0.size, zt.size))

    # A distribution over all possible final states is determined for each requested initial state
    for idx1, _z0 in enumerate(z0):
        # Determine the mean of the distribution
        for meth in solver_methods:
            fluid_path = int.solve_ivp(lambda _, z:
                                       b_rate(z, param) - d_rate(z, param),
                                       [t_linspace[0],
                                        t_linspace[-1] + 1e-100],
                                       [_z0],
                                       t_eval=t_linspace,
                                       method=meth)
            if fluid_path.success:
                break

        fluid_path = np.array(fluid_path.y[0])

        # The next few steps determine the covariance. See Section 2.3 of reference [1] for info
        m_function = np.exp(ut.trap_int(h_fun(fluid_path, param),
                                        t_linspace))

        integrand = np.divide(np.add(b_rate(fluid_path, param),
                                     d_rate(fluid_path, param)),
                              np.square(m_function))

        integral = ut.trap_int(integrand, t_linspace)

        indices = np.where(np.in1d(t_linspace, t))[0]

        m = fluid_path[indices]

        s2 = np.multiply(np.square(m_function[indices]),
                         integral[indices])

        # Given the mean and covariance, use a normal distribution pdf to determine probabilities
        # for requested terminal states
        for idx2 in range(zt.size):
            if t.size == 1:
                output[idx1, idx2] = norm.pdf(zt[idx2], loc=m[0],
                                              scale=np.sqrt(s2[0]))
            else:
                for idx3 in range(t.size):
                    output[idx3, idx1, idx2] = \
                        norm.pdf(zt[idx2], loc=m[idx3], scale=np.sqrt(s2[idx3]))
    # Numerical overflow is possible (particularly for stiff odes when determining 'fluid_path'),
    # which result in nan values. Provide a warning about this.
    if np.isnan(output).any():
        warnings.warn("Failed to compute some probabilities, some values "
                      "have been returned as nan. ",
                      category=RuntimeWarning)
    return output



import numpy as np
import birdepy.aug_num as _aug_num
import birdepy.aug_ilt as _aug_ilt
import birdepy.aug_expm as _aug_expm
import birdepy.utility as ut
import mpmath as mp


def f_fun_bld(sorted_data, param, b_rate, d_rate, likelihood, technique,
              idx_known_p, known_p, model, z_trunc, j_tol, h_tol, options):
    """
    Returns the surrogate likelihood function for use in the
    expectation-maximization algorithm. This function is defined inline in Section 2.1
    of reference [1].

    This function is used by the five em_*.py modules where * is cg, Lange, none,
    qn1 or qn2.

    [1] Jamshidian, M., & Jennrich, R. I. (1997). Acceleration of the EM algorithm
     by using quasi‐Newton methods. Journal of the Royal Statistical Society:
     Series B (Statistical Methodology), 59(3), 569-587.
    """
    param = ut.p_bld(param, idx_known_p, known_p)
    if technique == 'num':
        aug_data = _aug_num.aug_bld_num(sorted_data, param, b_rate, d_rate,
                                        likelihood, model, z_trunc, j_tol,
                                        h_tol, options)

    elif technique == 'ilt':
        if 'laplace_method' in options.keys():
            laplace_method = options['laplace_method']
        else:
            laplace_method = 'cme-talbot'
        if 'lentz_eps' in options.keys():
            lentz_eps = options['lentz_eps']
        else:
            lentz_eps = 1e-6
            options['lentz_eps'] = lentz_eps
        if 'k' in options.keys():
            k = options['k']
        else:
            k = 25
            options['k'] = k
        if 'precision' in options.keys():
            mp.dps = options['precision']
        else:
            mp.dps = 100
        aug_data = _aug_ilt.aug_bld_ilt(sorted_data, param, b_rate, d_rate,
                                        likelihood, model, z_trunc, j_tol,
                                        h_tol, lentz_eps, laplace_method, k,
                                        options)

    elif technique == 'expm':
        aug_data = _aug_expm.aug_bld_expm(sorted_data, param, b_rate, d_rate,
                                          likelihood, model, z_trunc, j_tol,
                                          h_tol, options)
    else:
        raise TypeError("Argument technique has an unknown value.")

    def f_fun(p_prime_prop):
        p_prime_prop = ut.p_bld(p_prime_prop, idx_known_p, known_p)
        output = 0
        for i in aug_data:
            if b_rate(i, p_prime_prop) > 0:
                output += aug_data[i][0] * np.log(b_rate(i, p_prime_prop))
            if d_rate(i, p_prime_prop) > 0:
                output += aug_data[i][1] * np.log(d_rate(i, p_prime_prop))
            output -= aug_data[i][2] * (b_rate(i, p_prime_prop) +
                                        d_rate(i, p_prime_prop))

        return output

    return f_fun, aug_data



import numpy as np
import mpmath as mp
from scipy.special import binom


def beta1(lam, mu, t):
    """
    Computes an important part of the transition probability function for a linear
    birth-and-death process. Corresponds to $\alpha$ defined below Equation (1)
    in reference [1]. Switches into a high precision mode if an exception is
    detected using machine precision.

    References
    ----------
    .. [1] Davison, A. C., Hautphenne, S., & Kraus, A. (2021). Parameter
    estimation for discretely observed linear birth‐and‐death processes.
    Biometrics, 77(1), 186-196.
    """
    if lam == mu:
        return lam * t / (1 + lam * t)
    else:
        with np.errstate(divide='raise', over='raise', under='raise',
                         invalid='raise'):
            try:
                output = mu * (np.exp((lam - mu) * t) - 1) \
                         / (lam * np.exp((lam - mu) * t) - mu)
                if output == 1:
                    raise Exception
            except:
                output = mp.fdiv(mp.fmul(mu,
                                         mp.fsub(mp.exp((mp.fmul(mp.fsub(lam,
                                                                         mu),
                                                                 t))),
                                                 1.0)),
                                 mp.fmul(lam,
                                         mp.fsub(mp.exp((mp.fmul(mp.fsub(lam,
                                                                         mu),
                                                                 t))),
                                                 mu)))
        return output


def beta2(lam, mu, t):
    """
    Computes an important part of the transition probability function for a linear
    birth-and-death process. Corresponds to $\beta$ defined below Equation (1)
    in reference [1]. Switches into a high precision mode if an exception is
    detected using machine precision.

    References
    ----------
    .. [1] Davison, A. C., Hautphenne, S., & Kraus, A. (2021). Parameter
    estimation for discretely observed linear birth‐and‐death processes.
    Biometrics, 77(1), 186-196.
    """
    if lam == mu:
        return beta1(lam, mu, t)
    else:
        with np.errstate(divide='raise', over='raise', under='raise',
                         invalid='raise'):
            try:
                output = lam * beta1(lam, mu, t) / mu
                if output == 1:
                    raise Exception
            except:
                output = mp.fmul(lam, mp.fdiv(beta1(lam, mu, t), mu))
        return output


def p_lin(z0, zt, lam, mu, t):
    """
    Computes the transition probability function for a linear
    birth-and-death process, as given by Equation (2) in reference [1].
    Switches into a high precision mode if an exception is
    detected using machine precision.

    References
    ----------
    .. [1] Davison, A. C., Hautphenne, S., & Kraus, A. (2021). Parameter
    estimation for discretely observed linear birth‐and‐death processes.
    Biometrics, 77(1), 186-196.
    """
    if zt == 0:
        with np.errstate(divide='raise', over='raise', under='raise',
                         invalid='raise'):
            try:
                pr = beta1(lam, mu, t) ** z0
            except:
                pr = float(mp.power(beta1(lam, mu, t), z0))
    else:
        pr = 0
        for j in np.arange(np.maximum(0, z0 - zt), z0, 1):
            with np.errstate(divide='raise', over='raise', under='raise',
                             invalid='raise'):
                try:
                    pr_j = binom(z0, j) * \
                           binom(zt - 1, z0 - j - 1) * \
                           (beta1(lam, mu, t) ** j) * \
                           (((1 - beta1(lam, mu, t)) * (1 - beta2(lam, mu, t)))
                            ** (z0 - j)) * \
                           (beta2(lam, mu, t) ** (zt - z0 + j))
                except:
                    pr_j = mp.fprod((mp.binomial(z0, j),
                                     mp.binomial(zt - 1.0, z0 - j - 1.0),
                                     mp.power(beta1(lam, mu, t), j),
                                     mp.power(mp.fmul(mp.fsub(1.0,
                                                              beta1(lam, mu, t)
                                                              ),
                                                      mp.fsub(1.0,
                                                              beta2(lam, mu,
                                                                    t))),
                                              z0 - j),
                                     mp.power(beta2(lam, mu, t),
                                              zt - z0 + j)))
            pr += float(pr_j)
    return pr



import numpy as np
import birdepy.utility as ut


def discrete(param, model, z0, times, k=1, method='exact', tau=0.1,
             survival=False, seed=None, display=False, **options):
    """Simulation of continuous-time birth-and-death processes at discrete
    observation times.

    Parameters
    ----------
    param : array_like
        The parameters governing the evolution of the birth-and-death
        process to be simulated.
        Array of real elements of size (m,), where ‘m’ is the number of
        parameters.
        These must be in the order given :ref:`here
        <Birth-and-death Processes>`).

    model : string
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst'
            - 'Ricker'
            - 'Hassell'
            - 'MS-S'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'M/M/1'
            - 'M/M/inf'
            - 'loss-system'
            - 'custom'

         If set to 'custom', then kwargs `b_rate` and `d_rate` must also be
         specified. See :ref:`here <Custom Models>` for more information.

    z0: int or callable
        The initial population size for each sample path.
        If it is a callable it should be a function that has no arguments and
        returns an int:

         ``z0() -> int``

    times : array_like
        The times at which the simulated birth-and-death is observed.
        Array of real elements of size (n,), where ‘n’ is the number of
        observation times.

    k : int, optional
        The number of sample paths to be simulated.

    method : string, optional
        Simulation algorithm used to generate samples (see
        :ref:`here<Simulation Algorithms>`). Should be one of:

            - 'exact' (default)
            - 'ea'
            - 'ma'
            - 'gwa'

    tau : scalar, optional
        Time between samples for the approximation methods 'ea', 'ma' and 'gwa'.

    survival : bool, optional
        If set to True, then the simulated sample paths are conditioned to
        have a positive population size at the final observation time.  Since This
        can greatly increase computation time.

    seed : int, Generator, optional
        If *seed* is not specified the random numbers are generated according
        to *np.random.default_rng()*. If *seed* is an *int*, random numbers are
        generated according to *np.random.default_rng(seed)*. If seed is a
        *Generator*, then that object is used. See
        :ref:`here <Reproducibility>` for more information.

    display : bool, optional
        If set to True, then a progress indicator is printed as the simulation
        is performed.

    Return
    -------
    out : array_like
        If k=1 a list containing sampled population size observations at
        `times`, generated according to `model`.
        Or if k>1, a numpy.ndarray containing k sample paths, each
        contained in a row of the array.

    Examples
    --------
    Simulating a unit rate Poisson process with observations at times
    [0, 1, 2, 3, 4, 5]: ::

        import birdepy as bd
        bd.simulate.discrete(1, 'Poisson', 0, times=[0, 1, 3, 4, 5])
        [0, 1, 3, 5, 5]

    Notes
    -----
    If you use this function for published work, then please cite [1].

    Sample paths are generated using a discrete-event simulation algorithm.
    See, for example, Algorithm 5.8 in [2].

    For a text book treatment on the theory of birth-and-death processes
    see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    :func:`birdepy.gpu_functions.probability()`  :func:`birdepy.gpu_functions.discrete()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Kroese, D.P., Taimre, T. and Botev, Z.I. (2013) Handbook of Monte
     Carlo methods. John Wiley & Sons.

    .. [3] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """
    # Convert parameters into a numpy array
    param = np.array(param)

    if len(param.shape) == 0:
        # In case param is specified as a scalar
        param = np.array([param])
    elif len(param.shape) > 1:
        raise TypeError("Argument `param` has an unsupported shape.")

    # Convert predetermined observation times into a numpy array
    times = np.array(times)

    # Set the random number generator
    if seed is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(seed)

    # Define the model-dependent birth and death rate functions
    if model == 'custom':
        b_rate = options['b_rate']
        d_rate = options['d_rate']
    else:
        b_rate = ut.higher_birth(model)
        d_rate = ut.higher_death(model)

    # Call a different function depending on the simulation method requested
    if method == 'exact':
        return discrete_exact(param, z0, times, k, survival, rng, display, b_rate, d_rate)
    elif method == 'ea':
        return discrete_ea(param, z0, times, k, survival, rng, display, b_rate, d_rate, tau)
    elif method == 'ma':
        return discrete_ma(param, z0, times, k, survival, rng, display, b_rate, d_rate, tau)
    elif method == 'gwa':
        return discrete_gwa(param, z0, times, k, survival, rng, display, b_rate, d_rate, tau)
    else:
        raise Exception(f"Argument `method` has unknown value '{method}'.")


def discrete_exact(param, z0, times, k, survival, rng, display, b_rate, d_rate):
    """Exact simulation of continuous-time birth-and-death processes at discrete
    observation times.
    """
    # Initialize array to store samples in
    sample = np.empty((k, times.shape[0]), dtype=np.int64)
    for sample_path in range(k):
        # This while loop has a break condition below governed by the 'survival' argument
        while True:
            # Generate the initial population
            if callable(z0):
                pop = z0()
            else:
                pop = z0
            rate = b_rate(pop, param) + d_rate(pop, param)
            # Determine the first event time >0
            if rate == 0:
                next_event_time = np.inf
            else:
                next_event_time = rng.exponential(1 / rate)
            for idx, next_observation_time in enumerate(times):
                # Update sample path up to an event time that exceeds the next observation time
                while next_event_time <= next_observation_time:
                    if (rng.uniform(0, 1) * rate) <= b_rate(pop, param):
                        pop += 1
                    else:
                        pop -= 1
                    rate = b_rate(pop, param) + d_rate(pop, param)
                    if rate > 0:
                        next_event_time += rng.exponential(1 / rate)
                    else:
                        next_event_time = np.inf
                sample[sample_path, idx] = pop
            if (survival and sample[sample_path, -1] > 0) or not survival:
                # This conditions on survival of the population for the
                # monitoring period
                break
        # Print progress if requested
        if display:
            print(100 * (sample_path + 1) / k, '% Complete')
    # Return output as a list if only one sample is requested
    if k == 1:
        out = sample[0].tolist()
    else:
        out = sample
    return out


def discrete_ea(param, z0, times, k, survival, rng, display, b_rate, d_rate, tau):
    """Approximate simulation of continuous-time birth-and-death processes at discrete
    observation times. Uses tau-leaping (or piecewise approximation) with rates fixed
    at the beginning of each subinterval (i.e., an Euler approximation scheme).
    """
    # Initialize array to store samples in
    sample = np.empty((k, times.shape[0]), dtype=np.int64)
    for sample_path in range(k):
        # This while loop has a break condition below governed by the 'survival' argument
        while True:
            # Generate the initial population
            if callable(z0):
                pop = z0()
            else:
                pop = z0
            # Determine the first event time >0
            next_event_time = tau
            for idx, next_observation_time in enumerate(times):
                # Update sample path up to an event time that exceeds the next observation time
                while next_event_time <= next_observation_time:
                    pop += rng.poisson(b_rate(pop, param) * tau)
                    pop -= rng.poisson(d_rate(pop, param) * tau)
                    if pop < 0:
                        pop = 0
                    next_event_time += tau
                sample[sample_path, idx] = pop
            if (survival and sample[sample_path, -1] > 0) or not survival:
                # This conditions on survival of the population for the
                # monitoring period
                break
        # Print progress if requested
        if display:
            print(100 * (sample_path + 1) / k, '% Complete')
    # Return output as a list if only one sample is requested
    if k == 1:
        out = sample[0].tolist()
    else:
        out = sample
    return out


def discrete_ma(param, z0, times, k, survival, rng, display, b_rate, d_rate, tau):
    """Approximate simulation of continuous-time birth-and-death processes at discrete
    observation times. Uses tau-leaping (or piecewise approximation) with rates fixed
    at an estimate of the expected value of the population size at the midpoint (in time)
    of each subinterval (i.e., a midpoint approximation scheme).
    """
    # Initialize array to store samples in
    sample = np.empty((k, times.shape[0]), dtype=np.int64)
    for sample_path in range(k):
        # This while loop has a break condition below governed by the 'survival' argument
        while True:
            # Generate the initial population
            if callable(z0):
                pop = z0()
            else:
                pop = z0
            # Determine the first event time >0
            next_event_time = tau
            for idx, next_observation_time in enumerate(times):
                # Update sample path up to an event time that exceeds the next observation time
                while next_event_time <= next_observation_time:
                    pop += rng.poisson(
                        b_rate((pop + 0.5 * tau * (b_rate(pop, param) -
                                                   d_rate(pop, param))), param) * tau)
                    pop -= rng.poisson(
                        d_rate((pop + 0.5 * tau * (b_rate(pop, param) -
                                                   d_rate(pop, param))), param) * tau)
                    if pop < 0:
                        pop = 0
                    next_event_time += tau
                sample[sample_path, idx] = pop
            if (survival and sample[sample_path, -1] > 0) or not survival:
                # This conditions on survival of the population for the
                # monitoring period
                break
        # Print progress if requested
        if display:
            print(100 * (sample_path + 1) / k, '% Complete')
    # Return output as a list if only one sample is requested
    if k == 1:
        out = sample[0].tolist()
    else:
        out = sample
    return out


def discrete_gwa(param, z0, times, k, survival, rng, display, b_rate, d_rate, tau):
    """Approximate simulation of continuous-time birth-and-death processes at discrete
    observation times. Uses a piecewise linear approximation where the rates are fixed
    at the start of each subinterval and then the process is assumed to evolve linearly.
    """
    # Initialize array to store samples in
    sample = np.empty((k, times.shape[0]), dtype=np.int64)
    for sample_path in range(k):
        # This while loop has a break condition below governed by the 'survival' argument
        while True:
            # Generate the initial population
            if callable(z0):
                pop = z0()
            else:
                pop = z0
            # Determine the first event time >0
            next_event_time = tau
            for idx, next_observation_time in enumerate(times):
                # Update sample path up to an event time that exceeds the next observation time
                while next_event_time <= next_observation_time:
                    lam = b_rate(pop, param) / pop if pop > 0 else 0
                    mu = d_rate(pop, param) / pop if pop > 0 else 0
                    p = 1 - beta1(lam, mu, tau)
                    number_survivors = rng.binomial(pop, p)
                    if number_survivors > 0:
                        pop = number_survivors + \
                              rng.negative_binomial(number_survivors,
                                                    1-beta2(lam, mu, tau))
                    else:
                        pop = 0
                    next_event_time += tau
                sample[sample_path, idx] = pop
            if (survival and sample[sample_path, -1] > 0) or not survival:
                # This conditions on survival of the population for the
                # monitoring period
                break
        # Print progress if requested
        if display:
            print(100 * (sample_path + 1) / k, '% Complete')
    # Return output as a list if only one sample is requested
    if k == 1:
        out = sample[0].tolist()
    else:
        out = sample
    return out


def beta1(lam, mu, t):
    """
    Computes an important part of the transition probability function for a linear
    linear-and-death process. Corresponds to $\alpha$ defined below Equation (1)
    in reference [1].

    References
    ----------
    .. [1] Davison, A. C., Hautphenne, S., & Kraus, A. (2021). Parameter
    estimation for discretely observed linear birth‐and‐death processes.
    Biometrics, 77(1), 186-196.
    """
    if lam == mu:
        return lam * t / (1 + lam * t)
    else:
        return mu * (np.exp((lam - mu) * t) - 1) / (lam * np.exp((lam - mu) * t) - mu)


def beta2(lam, mu, t):
    """
    Computes an important part of the transition probability function for a linear
    linear-and-death process. Corresponds to $\beta$ defined below Equation (1)
    in reference [1].

    References
    ----------
    .. [1] Davison, A. C., Hautphenne, S., & Kraus, A. (2021). Parameter
    estimation for discretely observed linear birth‐and‐death processes.
    Biometrics, 77(1), 186-196.
    """
    if lam == mu:
        return lam * t / (1 + lam * t)
    else:
        return lam * beta1(lam, mu, t) / mu


def continuous(param, model, z0, t_max, k=1, survival=False, seed=None,
               **options):
    """Simulation of continuous-time birth-and-death processes at birth and
    death event times.

    Parameters
    ----------
    param : array_like
        The parameters governing the evolution of the birth-and-death
        process to be simulated.
        Array of real elements of size (n,), where ‘n’ is the number of
        param.
        These must be in the order given :ref:`here
        <Birth-and-death Processes>`).


    model : string
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst'
            - 'Ricker'
            - 'Beverton-Holt'
            - 'Hassell'
            - 'MS-S'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'M/M/1'
            - 'M/M/inf'
            - 'loss-system'
            - 'custom'

         If set to 'custom', then kwargs `b_rate` and `d_rate` must also be
         specified. See :ref:`here <Custom Models>` for more information.

    z0: int or callable
        The initial population size for each sample path.
        If it is a callable it should be a function that has no arguments and
        returns an int:

         ``z0() -> int``

    t_max : scalar
        The simulation horizon. All events up to and including this time are
        included in the output.

    k : int, optional
        The number of sample paths to be simulated.

    survival : bool, optional
        If set to True, then the simulated sample paths are conditioned to
        have a positive population size at the final observation time. This
        can greatly increase computation time.

    seed : int, numpy.random._generator.Generator, optional
        If *seed* is not specified the random numbers are generated according
        to *np.random.default_rng()*. If *seed* is an *int*, random numbers are
        generated according to *np.random.default_rng(seed)*. If seed is a
        *Generator*, then that object is used. See
        :ref:`here <Reproducibility>` for more information.

    Returns
    -------
    jump_times : list
        If n=1 a list containing jump times, generated according to `model`
        or according to a birth-and-death process evolving according to
        `b_rate` and `d_rate`.
        Or if n>= 1, a list of lists where each list corresponds to the jump
        times from one sample path.

    pop_sizes : list
        If k=1 a list containing population sizes at the corresponding jump
        times, generated according to `model`.
        Or if k>1, a list of lists where each list corresponds to the
        population sizes corresponding to jump times from one sample path.

    Examples
    --------
    Simulating a unit rate Poisson process up to a t_max of 5: ::

        import birdepy as bd
        jump_times, pop_sizes = bd.simulate.continuous(1,'Poisson', 0, t_max=5)
        print(jump_times)
        print(pop_sizes)

    Outputs: ::

        [0, 0.0664050052043501, 0.48462937097695785, 2.2065719224651157]
        [0, 1, 2, 3]

    Notes
    -----
    If you use this function for published work, then please cite [1].

    Sample paths are generated using a discrete-event simulation algorithm.
    See, for example, Algorithm 5.8 in [2].

    For a text book treatment on the theory of birth-and-death processes
    see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Kroese, D.P., Taimre, T. and Botev, Z.I. (2013) Handbook of Monte
     Carlo methods. John Wiley & Sons.

    .. [3] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """
    # Convert parameters into a numpy array
    param = np.array(param)

    if len(param.shape) == 0:
        # In case param is specified as a scalar
        param = np.array([param])
    elif len(param.shape) > 1:
        raise TypeError("Argument `param` has an unsupported shape.")

    # Set the random number generator
    if seed is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(seed)

    # Define the model-dependent birth and death rate functions
    if model == 'custom':
        b_rate = options['b_rate']
        d_rate = options['d_rate']
    else:
        b_rate = ut.higher_birth(model)
        d_rate = ut.higher_death(model)

    # Initialize lists to store population size and jump time sample paths in.
    pop_sizes = []
    jump_times = []
    for sample_path in range(k):
        # This while loop has a break condition below governed by the 'survival' argument.
        while True:
            # Generate the initial population (which is assumed observed at time 0).
            if callable(z0):
                pop = z0()
            else:
                pop = z0
            # Initialize lists to store this sample path in with initial population and initial
            # observation time (which is always 0)
            _pop_sample = [pop]
            _jump_times = [0]

            # Determine the time of the first event, which is needed to determine whether to enter
            # the while loop below.
            rate = b_rate(pop, param) + d_rate(pop, param)
            if rate > 0:
                next_event_time = rng.exponential(1 / rate)
            else:
                next_event_time = np.inf

            # Update population record while the next event is less than or equal to simulation
            # horizon
            while next_event_time <= t_max:
                _jump_times.append(next_event_time)
                # Determine direction of jump
                if (rng.uniform(0, 1) * rate) <= b_rate(pop, param):
                    pop += 1
                else:
                    pop -= 1
                # Determine next event time
                rate = b_rate(pop, param) + d_rate(pop, param)
                if rate > 0:
                    next_event_time += rng.exponential(1 / rate)
                else:
                    next_event_time = np.inf
                _pop_sample.append(pop)
            if (survival and pop > 0) or not survival:
                # This conditions on survival of the population for the
                # monitoring period
                break
        pop_sizes.append(_pop_sample)
        jump_times.append(_jump_times)
    # Output a list if only one sample is requested and otherwise output a list of lists
    if k == 1:
        jump_times = jump_times[0]
        pop_sizes = pop_sizes[0]
    return jump_times, pop_sizes



import numpy as np
import warnings
from birdepy.interface_abc import discrete_est_abc
from birdepy.interface_dnm import discrete_est_dnm
from birdepy.interface_dnm import continuous_est_dnm
from birdepy.interface_lse import discrete_est_lse
from birdepy.interface_em import discrete_est_em
import birdepy.utility as ut
import time
from scipy.optimize import OptimizeResult
from birdepy.interface_aug import f_fun_bld as f_fun_bld
from birdepy import simulate


def estimate(t_data, p_data, p0, p_bounds, framework='dnm', model='Verhulst',
             scheme='discrete', con=(), known_p=(), idx_known_p=(),
             se_type='asymptotic', seed=None, ci_plot=False, export=False, display=False,
             **options):
    """Parameter estimation for (continuously or discretely observed)
    continuous-time birth-and-death processes.

    Parameters
    ----------
    t_data : list
        Observation times of birth-and-death process. If one trajectory is
        observed, then this is a list. If multiple trajectories are observed,
        then this is a list of lists where each list corresponds to a
        trajectory.

    p_data : list
        Observed populations of birth-and-death process at times in argument
        `t_data`. If one trajectory is observed, then this is a list. If
        multiple trajectories are observed, then this is a list of lists
        where each list corresponds to a trajectory.

    p0 : array_like
        Initial parameter guess. Array of real elements of size (n,), where n
        is the number of unknown parameters.

    p_bounds : list
        Bounds on parameters. Should be specified as a sequence of
        ``(min, max)`` pairs for each unknown parameter. See :ref:`here <Parameter Constraints>`.

    framework : string, optional
        Parameter estimation framework. Should be one of:

            - 'abc' (see :ref:`here <birdepy.estimate(framework='abc')>`)
            - 'dnm' (default) (see :ref:`here <birdepy.estimate(framework='dnm')>`)
            - 'em' (see :ref:`here <birdepy.estimate(framework='em')>`)
            - 'lse' (see :ref:`here <birdepy.estimate(framework='lse')>`)

        Additional kwargs are available for each framework.

    model : string, optional
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst' (default)
            - 'Ricker'
            - 'Hassell'
            - 'MS-S'
            - 'Moran'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'M/M/1'
            - 'M/M/inf'
            - 'loss-system'
            - 'custom'

         If set to 'custom', then kwargs `b_rate` and `d_rate` must also be
         specified. See :ref:`here <Custom Models>` for more information.

    scheme : string, optional
        Observation scheme. Should be one of:

            - 'discrete' (default)
            - 'continuous' (see :ref:`here <birdepy.estimate(scheme='continuous')>`)

        If set to 'continuous', then it is assumed that the population is
        observed continuously with jumps occuring at times in `t_data` into
        corresponding states in `p_data`.

    con : {Constraint, dict} or List of {Constraint, dict}, optional
        Constraints definition for parameters (only for kwarg `opt_method`
        equal to 'differential-evolution', 'COBYLA', 'SLSQP' or
        'trust-constr').   See :ref:`here <Parameter Constraints>` for more
        information.

    known_p : array_like, optional
        List of known parameter values. For built in models these must be in
        their canonical order as given (:ref:`here <Birth-and-death Processes>`). If this
        argument is given, then argument `idx_known_p` must also be specified.
        See :ref:`here <Known Parameters>` for more information.

    idx_known_p : array_like, optional
        List of indices of known parameters (as given in argument 'known_p').
        For built in models indices must correspond to canonical order as given
        :ref:`here <Birth-and-death Processes>`. If this argument is given, then argument
        `known_p` must also be specified. See :ref:`here <Known Parameters>`
        for more information.

    se_type: string, optional
        Should be one of: 'none' (default), 'simulated', or 'asymptotic'.
        See :ref:`here <Confidence Regions>` for more information.

    display : bool, optional
        If True, then progress updates are provided for some methods.

    ci_plot : bool, optional
        Enables confidence region plotting for 2d parameter estimates.

    export : str, optional
        File name for export of confidence region figure to a LaTeX file.

    Returns
    -------
    res : EstimationOutput
        The estimation output represented as an :func:`EstimationOutput`
        object. Important attributes are: `p` the parameter estimate, `se` the
        standard error estimate, `cov` the estimated covariance of the
        assumed distribution of the parameter estimate, `val` is the log-likelihood
        for 'framework' `dnm' and 'em', squared error for 'framework' `lse',
        `capacity` is the estimated carrying capacity.

    Examples
    --------
    Example 1: Simulate a discretely observed sample path and estimate the parameters using the
    alternative frameworks.
    First simulate some sample paths of a Ricker model using
    :func:`birdepy.simulate.discrete()`: ::

        import birdepy as bd
        t_data = list(range(100))
        p_data = bd.simulate.discrete([0.75, 0.25, 0.02, 1], 'Ricker', 10, t_data,
                                      survival=True, seed=2021)

    Then, using the simulated data, estimate the parameters using 'dnm', 'lse'
    and 'em' as the argument of `framework`: ::

        est_dnm = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[0,1], [0,1], [0, 0.1]],
                              framework='dnm', model='Ricker', idx_known_p=[3], known_p=[1])
        est_em = bd.estimate(t_data, p_data, [1, 0.5, 0.05], [[0,1], [0,1], [1e-6,0.1]],
                              framework='em', model='Ricker', idx_known_p=[3], known_p=[1])
        est_abc = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[0,1], [0,1], [0, 0.1]],
                              framework='abc', model='Ricker', idx_known_p=[3], known_p=[1])
        est_lse = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[0,1], [0,1], [0, 0.1]],
                              framework='lse', model='Ricker', idx_known_p=[3], known_p=[1], se_type='simulated')
        print(f'dnm estimate: {est_dnm.p}, dnm standard errors: {est_dnm.se}')
        print(f'lse estimate: {est_lse.p}, lse standard errors: {est_lse.se}')
        print(f'abc estimate: {est_abc.p}, abc standard errors: {est_abc.se}')
        print(f'em estimate: {est_em.p}, em standard errors: {est_em.se}')

    Outputs: ::

        dnm estimate: [0.7477212189824904, 0.2150484536334751, 0.022745124483227304] , dnm standard errors: [0.16904225 0.03443199 0.00433567]
        em estimate: [0.7375802511179848, 0.19413965548145604, 0.024402343633644553] , em standard errors: [0.15742852 0.02917437 0.00429763]
        abc estimate: [0.6318632898413052, 0.02074882329749562, 0.06580340596326038], abc standard errors: [0.22865334, 0.0148124, 0.0129306]
        lse estimate: [0.7941741586214265, 0.2767698457541133, 0.01935636627568731] , lse standard errors: [0.1568291  0.19470746 0.01243208]

    Alternatively, we may be interested in continuously observed data.

    Example 2: Simulate a continuous sample path and estimate the parameters.

    Simulate some synthetic data: ::

        t_data, p_data = bd.simulate.continuous([0.75, 0.25, 0.02, 1], 'Ricker', 10,
                                                100, survival=True, seed=2021)

    Estimate: ::

        est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[0,1], [0,1], [0, 0.1]],
                          model='Ricker', idx_known_p=[3], known_p=[1], scheme='continuous')
        print(est.p)

    Outputs: ::

        [0.7603171062895576, 0.2514810854871476, 0.020294342655751033]

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    If `model` is 'Poisson', then the true dnm is immediately returned (i.e.,
    the total number of events during the observation periods divided by the
    total observation time).

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    :func:`birdepy.gpu_functions.probability()`  :func:`birdepy.gpu_functions.discrete()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    # Start a timer so that the total time to run the function can be returned
    tic = time.time()

    # Key word arguments can be passed as a dictionary which is itself the value
    # of a key word argument with key 'options', in this case the dictionary
    # must be extracted:
    if 'options' in options.keys():
        options = options['options']

    # Convert the arguments of 'p0', 'known_p' and 'idx_known_p' into numpy arrays
    p0 = np.array(p0)
    known_p = np.array(known_p)
    idx_known_p = np.array(idx_known_p)

    # Augment the options dictionary with items used by the optimization routines later
    options = ut.add_options(options)

    # Set the random number generator
    if seed is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(seed)

    # Define the model-dependent birth and death rate functions
    if model == 'custom':
        b_rate = options['b_rate']
        d_rate = options['d_rate']
    else:
        b_rate = ut.higher_birth(model)
        d_rate = ut.higher_death(model)

    # Create object to return results in
    class EstimationOutput(OptimizeResult):
        pass

    # Determine maximum and minimum population sizes (or "truncation" points) of process
    if 'z_trunc' in options.keys():
        z_trunc = options['z_trunc']
    else:
        z_obs_min = np.inf
        z_obs_max = 0
        for idx in range(len(p_data)):
            z_obs_min = np.minimum(z_obs_min, np.amin(p_data[idx]))
            z_obs_max = np.maximum(z_obs_max, np.amax(p_data[idx]))
        obs_range = z_obs_max - z_obs_min
        z_trunc = [int(np.maximum(0, z_obs_min - 2 * obs_range)),
                   int(z_obs_max + 2 * obs_range)]
        options['z_trunc'] = z_trunc

    # Continuously and discretely observed data each have different parameter estimation methods.
    # This control structure diverts to the continuous observation estimation dnm module for
    # continuous data or into another control structure that then diverts between the 4
    # ('abc', 'dnm', 'em' and 'lse) modules for discrete observation parameter estimation.
    if scheme == 'continuous':
        # Extract scheme specific key word arguments the dictionary 'options'
        # method.
        if 'opt_method' in options.keys():
            opt_method = options['opt_method']
        else:
            opt_method = 'L-BFGS-B'

        if framework != 'dnm':
            raise TypeError("Argument `framework` must be 'dnm' when "
                            "argument `scheme` is 'continuous'.")

        # Obtain estimate and associated covariance matrix
        opt, cov = continuous_est_dnm(t_data, p_data, p0, b_rate, d_rate,
                                      p_bounds, con, known_p, idx_known_p,
                                      opt_method, options)

        # Prepare items for output
        p_est = opt.x
        message = opt.message
        success = opt.success
        err = opt.fun
        iterations = 'Not applicable for `mle`. '
        accelerator = 'Not applicable for `mle`. '
        samples = 'Not applicable for `mle`. '
        method = 'mle'

    elif scheme == 'discrete':

        # Sort the data into a more efficient format
        data = ut.data_sort(t_data, p_data)

        # Each of the frameworks for obtaining parameter estimates are contained in their own
        # module interface_* where * depends on the framework. This control structure determines
        # which module to use depending on the value of 'framework'.  Framework specific key word
        # arguments are extracted from the dictionary 'options' and then used in the module
        # associated with the framework.
        if framework == 'abc':
            if 'k' in options.keys():
                k = options['k']
            else:
                k = 100
                options['k'] = k
            if 'eps_abc' in options.keys():
                eps_abc = options['eps_abc']
                if np.isscalar(eps_abc):
                    eps_abc = [eps_abc]
            else:
                eps_abc = 'dynamic'
                options['eps_abc'] = eps_abc
            if 'max_its' in options.keys():
                max_its = options['max_its']
            else:
                if eps_abc == 'dynamic':
                    max_its = 3
                else:
                    max_its = len(eps_abc)
                options['max_its'] = max_its
            if max_its != len(eps_abc) and eps_abc != 'dynamic':
                raise ValueError("Length of argument `abc_eps` "
                                 "does not match argument `its`.")
            if 'max_q' in options.keys():
                max_q = options['max_q']
            else:
                max_q = 0.99
                options['max_q'] = max_q
            if 'method' in options.keys():
                method = options['method']
            else:
                method = 'gwa'
                options['method'] = method
            if 'tau' in options.keys():
                tau = options['tau']
            else:
                tau = np.inf
                for i in data:
                    if i[2] < tau:
                        tau = i[2]
                tau = min(tau/10, 0.1)
                options['tau'] = tau
            if 'distance' in options.keys():
                distance = options['distance']
            else:
                distance = None
                options['distance'] = distance
            if 'stat' in options.keys():
                stat = options['stat']
            else:
                stat = 'mean'
                options['stat'] = stat
            if stat not in ['mean', 'median']:
                raise TypeError("Argument of 'stat' has an unknown value.")
            if 'gam' in options.keys():
                gam = options['gam']
            else:
                gam = 5
                options['gam'] = gam
            if 'eps_change' in options.keys():
                eps_change = options['eps_change']
            else:
                eps_change = 5
                options['eps_change'] = eps_change
            # Obtain estimate, associated covariance matrix, and accepted samples
            p_est, cov, samples = \
                discrete_est_abc(data, eps_abc, distance, stat, k, gam, max_its,
                                 max_q, eps_change, method, b_rate, d_rate,
                                 idx_known_p, known_p, p_bounds, con, tau, rng,
                                 display)

            message = 'Not applicable for `abc`.'
            success = 'Not applicable for `abc`.'
            iterations = 'Not applicable for `abc`. '
            err = 'Not applicable for `abc`.'

        elif framework == 'dnm':
            if 'likelihood' in options.keys():
                likelihood = options['likelihood']
            else:
                likelihood = 'expm'
                options['likelihood'] = 'expm'
            if 'opt_method' in options.keys():
                opt_method = options['opt_method']
            else:
                if con == ():
                    opt_method = 'L-BFGS-B'
                else:
                    opt_method = 'SLSQP'
                options['opt_method'] = opt_method

            # Obtain estimate and associated covariance matrix
            opt, cov = discrete_est_dnm(data, likelihood, model,
                                        z_trunc, idx_known_p, known_p, p0,
                                        p_bounds, con, opt_method, options)

            # Prepare items for output
            p_est = opt.x
            message = opt.message
            success = opt.success
            err = -opt.fun
            iterations = 'Not applicable for `dnm`. '
            method = 'Not applicable for `dnm`. '
            samples = 'Not applicable for `dnm`. '

        elif framework == 'em':
            if 'accelerator' in options.keys():
                accelerator = options['accelerator']
            else:
                accelerator = 'Lange'
            if 'opt_method' in options.keys():
                opt_method = options['opt_method']
            elif accelerator == 'none':
                opt_method = 'SLSQP'
            else:
                opt_method = 'SLSQP'
            if 'technique' in options.keys():
                technique = options['technique']
            else:
                technique = 'expm'
            if 'likelihood' in options.keys():
                likelihood = options['likelihood']
            elif technique == 'expm':
                likelihood = 'expm'
            elif technique == 'ilt':
                likelihood = 'ilt'
            else:
                likelihood = 'expm'
            if 'max_it' in options.keys():
                max_it = options['max_it']
            else:
                max_it = 25
            if 'i_tol' in options.keys():
                i_tol = options['i_tol']
            else:
                i_tol = 1e-3
            if 'j_tol' in options.keys():
                j_tol = options['j_tol']
            else:
                j_tol = 1e-2
            if 'h_tol' in options.keys():
                h_tol = options['h_tol']
            else:
                h_tol = 1e-2

            if accelerator == 'test_aug':
                # This can be used to return the augmented data for
                # parameter values p0 using bd.discrete_est, which may be useful
                # for performance evaluation or debugging.
                sorted_data = ut.data_sort_2(data)

                aug_data = f_fun_bld(sorted_data, p0, b_rate, d_rate, likelihood,
                                     technique, idx_known_p, known_p, model,
                                     z_trunc, j_tol, h_tol, options)[1]

                return aug_data

            # Obtain estimate, associated covariance matrix and per-iteration estimates
            p_est, cov, ll, iterations = discrete_est_em(
                data, p0, technique, accelerator, likelihood, p_bounds, con,
                known_p, idx_known_p, model, b_rate, d_rate, z_trunc, max_it,
                i_tol, j_tol, h_tol, display, opt_method, options)

            # Prepare items for output
            message = 'Not applicable for `em`.'
            success = 'Not applicable for `em`.'
            err = ll
            method = accelerator
            samples = 'Not applicable for `em`. '

        elif framework == 'lse':
            if 'opt_method' in options.keys():
                opt_method = options['opt_method']
            else:
                opt_method = 'L-BFGS-B'
                options['opt_method'] = opt_method
            if 'squares' in options.keys():
                squares = options['squares']
            else:
                squares = 'fm'
                options['squares'] = squares

            # Obtain estimate
            opt = discrete_est_lse(data, squares, model, b_rate, d_rate, z_trunc,
                                   idx_known_p, known_p, p0, p_bounds,
                                   con, opt_method, options)

            # Prepare items for output
            p_est = opt.x
            message = opt.message
            success = opt.success
            err = opt.fun
            cov = 'Not applicable for `lse`.'
            iterations = 'Not applicable for `lse`. '
            method = squares
            samples = 'Not applicable for `lse`. '

        else:
            raise ValueError("Argument `framework` has an unknown value. Should "
                             "be one of: 'abc', 'dnm', 'lse', or 'em'.")
    else:
        raise ValueError("Argument `scheme` has an unknown value. Should "
                         "be one of: 'discrete' or 'continuous'.")

    # Prepare for plotting
    if 'xlabel' in options.keys():
        xlabel = options['xlabel']
    else:
        xlabel = "$\\theta_1$"
    if 'ylabel' in options.keys():
        ylabel = options['ylabel']
    else:
        ylabel = "$\\theta_2$"

    # Compute confidence regions and standard errors, and plot confidence regions.
    # This control structure diverts between the 'asymptotic' and 'simulated'
    # approaches to doing these tasks.
    if se_type == 'asymptotic':
        if framework == 'lse':
            se = "Asymptotic confidence intervals are not available for " \
                 "`framework' 'lse'. Set argument `se_type` to 'none'" \
                 "or 'simulated'."
        else:
            # For frameworks 'abc', 'dnm' and 'em', use cov value computed above
            try:
                if cov.shape[0] == 1:
                    se = list(np.sqrt(np.diag(cov))[0])
                else:
                    se = list(np.sqrt(np.diag(cov)))
            except:
                se = 'Error computing standard errors. Covariance matrix may have' \
                     'negative diagonal entries.'
            if ci_plot:
                ut.confidence_region(mean=p_est, cov=cov, obs=None, se_type=se_type, xlabel=xlabel,
                                     ylabel=ylabel, export=export)
    elif se_type == 'simulated':
        if 'num_samples' in options.keys():
            num_samples = options['num_samples']
        else:
            num_samples = 100

        # Augment estimate of unknown parameters with known parameters
        param = ut.p_bld(p_est, idx_known_p, known_p)

        # This section generates synthetic data samples according to the estimate and then
        # performs the estimation using the synthetic data to obtain a collection of estimates.
        # The control structures are mostly to ensure the correct estimation procedure is used on
        # the synthetic data.
        bootstrap_samples = np.zeros((num_samples, p0.size))
        for idx in range(num_samples):
            if scheme == 'continuous':
                if type(t_data[0]) == list:
                    t_data_temp = []
                    p_data_temp = []
                    for idx in t_data:
                        times, pops = simulate.continuous(p_est,
                                                          model,
                                                          p_data[idx][0],
                                                          model,
                                                          t_data[idx][-1],
                                                          seed=rng)
                        t_data_temp.append(times)
                        p_data_temp.append(pops)
                else:
                    t_data_temp, p_data_temp = simulate.continuous(
                        p_est, model, p_data[idx][0], t_data[idx][-1])
                bootstrap_samples[idx, :] = continuous_est_dnm(
                    t_data_temp, p_data_temp, p0, b_rate, d_rate, p_bounds, con,
                    known_p, idx_known_p, opt_method, options)
            else:  # observation scheme is discrete
                temp_data = {}
                for i in data:
                    for _ in range(data[i]):
                        new_zt = simulate.discrete(param, model,
                                                   i[0], [0, i[2]], k=1,
                                                   seed=rng)[1]
                        new_point = (i[0], new_zt, i[2])
                        if new_point in temp_data:
                            temp_data[new_point] += 1
                        else:
                            temp_data[new_point] = 1
                if framework == 'dnm':
                    bootstrap_samples[idx, :] = discrete_est_dnm(
                        temp_data, likelihood, model, z_trunc, idx_known_p,
                        known_p, p_est, p_bounds, con, opt_method, options)[0].x
                elif framework == 'lse':
                    bootstrap_samples[idx, :] = discrete_est_lse(
                        temp_data, squares, model, b_rate, d_rate, z_trunc,
                        idx_known_p, known_p, p_est, p_bounds, con, opt_method,
                        options).x
                else:  # framework is 'em'
                    bootstrap_samples[idx, :] = discrete_est_em(
                        data, p_est, technique, accelerator, likelihood, p_bounds,
                        con, known_p, idx_known_p, model, b_rate, d_rate,
                        z_trunc, max_it, i_tol, j_tol, h_tol, display,
                        opt_method, options)[0]
            if display:
                print('Boostrap confidence region progress:',
                      100 * (idx + 1) / num_samples, '%')
            cov = np.cov(bootstrap_samples, rowvar=False)
        try:
            # This control structure ensures standard errors are returned in the correct format
            if cov.shape[0] == 1:
                se = list(np.sqrt(np.diag(cov))[0])
            else:
                se = list(np.sqrt(np.diag(cov)))
        except:
            se = 'Error computing standard errors. Covariance matrix may have ' \
                 'negative diagonal entries.'
        if ci_plot:
            try:
                ut.confidence_region(mean=p_est, cov=cov, se_type=se_type,
                                     obs=bootstrap_samples, xlabel=xlabel,
                                     ylabel=ylabel, export=export)
            except:
                warnings.warn("Error plotting confidence regions. Estimated "
                              "covariance matrix may have negative diagonal "
                              "entries.'",
                              category=RuntimeWarning)
    elif se_type == 'none':
        se = 'Not requested (see argument `se_type`).'
    else:
        raise TypeError("Argument `se_type` has an unknown value. Possible "
                        "options are 'none', 'asymptotic' and 'simulated' ")

    # Determine capacity attribute of the output
    if model != 'custom':
        # Build a function that returns fixed points of fluid model
        capacity_finder = ut.higher_zf_bld(model)
        # Use function to find fixed points in terms of estimate of unknown parameters augmented
        # with known parameters
        capacity = capacity_finder(ut.p_bld(np.array(p_est), idx_known_p, known_p))
    else:
        capacity = "Functionality not available for custom models."

    return EstimationOutput(p=list(p_est), capacity=capacity, val=err, cov=cov,
                            se=se, compute_time=time.time() - tic,
                            framework=framework, message=message,
                            success=success, iterations=iterations,
                            method=method, p0=list(p0),
                            scheme=scheme, samples=samples, model=model)



import numpy as np
import birdepy.utility as ut
from birdepy.interface_aug import f_fun_bld as f_fun_bld
from birdepy.interface_dnm import bld_ll_fun
import birdepy.utility_em as utem


def discrete_est_em_qn2(data, sorted_data, p0, likelihood, technique, known_p,
                        idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds,
                        con, max_it, i_tol, j_tol, h_tol, display, opt_method,
                        options):
    """
    Executes an expectation-maximization algorithm, which is accelerated using
    conjugate gradient ideas, to estimate parameters for a
    population-size-dependent birth-and-death process.

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.

    """
    #Initialize a list to return parameters in each iteration and store the initial parameters in it
    iterations = [p0]

    it = 1
    difference = i_tol + 1
    p_est = p0

    # Build the log-likelihood function assuming all parameters are known
    pre_ll = bld_ll_fun(data, likelihood, model, z_trunc, options)

    # Negate the surrogate log-likelihood function so that scipy.optimize.minimize()
    # Can be used to find a maximum
    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll(param)

    g = ut.Jacobian(ll, p_est, p_bounds)

    # Build surrogate log-likelihood function (defined in Section 2.1 of ref [1])
    f_fun = f_fun_bld(sorted_data, p_est, b_rate, d_rate, likelihood,
                      technique, idx_known_p, known_p, model, z_trunc,
                      j_tol, h_tol, options)[0]

    # Negate the surrogate log-likelihood function so that scipy.optimize.minimize()
    # Can be used to find a maximum
    def error_fun(p_prop):
        return -f_fun(p_prop)

    # Find a zero of an EM step (defined as $\tilde g(\theta) at start of Section 2.2
    # of reference [1])
    g_tilde = ut.minimize_(error_fun, p_est, p_bounds, con, opt_method,
                           options).x - p_est

    S = np.zeros((p0.size, p0.size))

    while difference > i_tol and it < max_it:

        # Step (a) of QN2 algorithm in reference [1]
        direction = g_tilde - np.matmul(S, g)

        # Step (b) of QN2 algorithm in reference [1]

        a_opt = utem.line_search_ll(direction, data, p_est, likelihood,
                                    known_p, idx_known_p, model, z_trunc,
                                    p_bounds, con, opt_method, options)[0]
        p_est_change = np.multiply(a_opt, direction)

        # First part of Step (c) of QN2 algorithm in reference [1]
        g_change = ut.Jacobian(ll, p_est + p_est_change, p_bounds) - g

        # Build surrogate log-likelihood function (defined in Section 2.1 of ref [1])
        f_fun = f_fun_bld(sorted_data, p_est + p_est_change, b_rate, d_rate,
                          likelihood, technique, idx_known_p, known_p, model,
                          z_trunc, j_tol, h_tol, options)[0]

        # Negate the surrogate log-likelihood function so that scipy.optimize.minimize()
        # Can be used to find a maximum
        def error_fun(p_prop):
            return -f_fun(p_prop)

        # Second part of Step (c) of QN2 algorithm in reference [1]
        g_tilde_change = ut.minimize_(error_fun, p_est, p_bounds, con,
                                      opt_method, options).x\
                         - p_est - p_est_change - g_tilde

        # First part of Step (d) of QN2 algorithm in reference [1]
        p_est_star_change = -g_tilde_change + np.matmul(S, g_change)

        # These are all part of Equation (4.1) in reference [1]
        s1 = (np.dot(g_change, p_est_star_change) /
              np.dot(g_change, p_est_change))
        s2 = np.dot(g_change, p_est_change)
        S1 = np.matmul(p_est_change, np.transpose(p_est_change))
        S2 = np.matmul(p_est_star_change, np.transpose(p_est_change))

        # Equation (4.1) (second part of Step (d) of QN2 algorithm) in reference [1]
        S_change = S1 * (1 + s1) / s2 - (S2 + np.transpose(S2)) / s2

        # Step (e) of QN2 algorithm in reference [1]
        g += g_change
        g_tilde += g_tilde_change
        S += S_change

        p_est += p_est_change

        # Compute the difference between this iteration estimate and
        # previous iteration estimate
        difference = np.sum(np.abs(p_est_change))
        # Store the estimate
        iterations.append(p_est)

        # Print a progress update if requested
        if display:
            print('Iteration ', it, ' estimate is: ', p_est)
        it += 1

    return list(p_est), np.array(iterations)



import numpy as np
import birdepy.utility_probability as ut
import warnings


def probability_gwa(z0, zt, t, param, b_rate, d_rate, anchor):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *Galton-Watson approximation* method.

    To use this function call :func:`birdepy.probability` with `method` set to
    'gwa'::

        birdepy.probability(z0, zt, t, param, method='gwa', anchor='midpoint')

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------
    anchor : string, optional
        Determines which state z is used to determine the linear approximation.
        Should be one of: 'initial' (z0 is used), 'midpoint' (default, 0.5*(z0+zt) is used),
        'terminal' (zt is used), 'max' (max(z0, zt) is used), or 'min' (min(z0, zt) is used).

    Examples
    --------
    Approximate transition probability for a Verhulst model using a linear approximation: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='gwa')[0][0]

    Outputs: ::

        0.00227651766770292

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """

    # If more than one time is requested it is easiest to divert into a different code block
    if t.size == 1:
        # Initialize an output array to be filled in as we loop over initial and terminal states
        output = np.zeros((z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            for idx2, _zt in enumerate(zt):
                # Set the rates of the linear approximation according the argument of 'anchor'
                if anchor == 'midpoint':
                    midpoint = 0.5 * (_z0 + _zt)
                    lam = b_rate(midpoint, param) / midpoint
                    mu = d_rate(midpoint, param) / midpoint
                elif anchor == 'initial':
                    lam = b_rate(_z0, param) / _z0
                    mu = d_rate(_z0, param) / _z0
                elif anchor == 'terminal':
                    lam = b_rate(_zt, param) / _zt
                    mu = d_rate(_zt, param) / _zt
                elif anchor == 'max':
                    max_z = max(_z0, _zt)
                    lam = b_rate(max_z, param) / max_z
                    mu = d_rate(max_z, param) / max_z
                elif anchor == 'min':
                    min_z = min(_z0, _zt)
                    lam = b_rate(min_z, param) / min_z
                    mu = d_rate(min_z, param) / min_z
                else:
                    raise TypeError(
                        "Argument 'anchor' has an unknown value. Should be one of 'midpoint'"
                        " 'initial', 'terminal', 'max' or 'min'.")
                # Determine the transition probability of the linear approximation
                pr = ut.p_lin(_z0, _zt, lam, mu, t[0])
                if not 0 <= pr <= 1.0:
                    # Warn if a probability outside [0,1] is computed
                    warnings.warn("Probability not in [0, 1] computed, "
                                  "some output has been replaced by a "
                                  "default value. "
                                  " Results may be unreliable.",
                                  category=RuntimeWarning)
                    # If probabilities are less than 0 we set them to 0 and if they are greater
                    # than 0 we set them to 1
                    if pr < 0:
                        pr = 0.0
                    else:
                        pr = 1.0
                # Fill in the output array with the computed probability
                output[idx1, idx2] = pr
    else:
        # Initialize an output array to be filled in as we loop over times, initial and terminal
        # states
        output = np.zeros((t.size, z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            for idx2, _zt in enumerate(zt):
                for idx3, _t in enumerate(t):
                    # Set the rates of the linear approximation according the argument of 'anchor'
                    if anchor == 'midpoint':
                        midpoint = 0.5 * (_z0 + _zt)
                        lam = b_rate(midpoint, param) / midpoint
                        mu = d_rate(midpoint, param) / midpoint
                    elif anchor == 'initial':
                        lam = b_rate(_z0, param) / _z0
                        mu = d_rate(_z0, param) / _z0
                    elif anchor == 'terminal':
                        lam = b_rate(_zt, param) / _zt
                        mu = d_rate(_zt, param) / _zt
                    elif anchor == 'max':
                        max_z = max(_z0, _zt)
                        lam = b_rate(max_z, param) / max_z
                        mu = d_rate(max_z, param) / max_z
                    elif anchor == 'min':
                        min_z = min(_z0, _zt)
                        lam = b_rate(min_z, param) / min_z
                        mu = d_rate(min_z, param) / min_z
                    else:
                        raise TypeError(
                            "Argument 'anchor' has an unknown value. Should be one of 'midpoint'"
                            " 'initial', 'terminal', 'max' or 'min'.")
                    # Determine the transition probability of the linear approximation
                    pr = ut.p_lin(_z0, _zt, lam, mu, t[0])
                    # Warn if a probability outside [0,1] is computed
                    if not 0 <= pr <= 1.0:
                        warnings.warn("Probability not in [0, 1] computed, "
                                      "some output has been replaced by a "
                                      "default value. "
                                      " Results may be unreliable.",
                                      category=RuntimeWarning)
                        # If probabilities are less than 0 we set them to 0 and if they are greater
                        # than 0 we set them to 1
                        if pr < 0:
                            pr = 0.0
                        else:
                            pr = 1.0
                    # Fill in the output array with the computed probability
                    output[idx3, idx1, idx2] = pr

    return output



import numpy as np
import birdepy.utility as ut
from scipy.stats import poisson


def p_mat_bld_uniform(q_mat, t, _k, num_states):
    """
    Builds and returns the discrete time Markov chain utilized by
    the uniformization method.
    """
    m = np.amax(np.absolute(np.diag(q_mat)))
    a_mat = np.divide(q_mat, m) + np.eye(num_states)
    w = np.eye(num_states)
    poisson_terms = poisson.pmf(range(_k + 1), m * t)
    p_mat = np.multiply(w, poisson_terms[0])
    for idx in np.arange(1, _k + 1, 1):
        w = np.matmul(w, a_mat)
        p_mat += np.multiply(w, poisson_terms[idx])

    return p_mat


def probability_uniform(z0, zt, t, param, b_rate, d_rate, z_trunc, k):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *uniformization* method.

    To use this function call :func:`birdepy.probability` with `method` set to
    'uniform'::

        birdepy.probability(z0, zt, t, param, method='uniform', k=1000, z_trunc=())

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------
    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default
        ``z_trunc=[z_min, z_max]`` where ``z_min=max(0, min(z0, zt) - 100)``
        and ``z_max=max(z0, zt) + 100``.

    k : int, optional
        Number of terms to include in approximation to probability.

    Examples
    --------
    Approximate transition probability for a Verhulst model using uniformization: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='uniform')[0][0]

    Outputs: ::

        0.002741422482539626

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    For more information on this method see [3] and [4].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Grassman, K. W. Transient solutions in Markovian queueing systems.
     Computers & Operations Research, 4(1):47-53, 1977.

    .. [4] van Dijk N.M., van Brummelen, S.P.J and Boucherie, R.J.
     Uniformization: Basics, extensions and applications. Performance
     Evaluation, 118:8-32, 2018.

    """
    # Extract the min and max population sizes considered
    z_min, z_max = z_trunc

    # Count how many states are considered
    num_states = int(z_max - z_min + 1)

    # If more than one time is requested it is easiest to divert into a different code block
    if t.size == 1:
        q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)
        p_mat = p_mat_bld_uniform(q_mat, t, k, num_states)
        output = p_mat[np.ix_(np.array(z0 - z_min, dtype=np.int32),
                              np.array(zt - z_min, dtype=np.int32))]
    else:
        # Initialize an output array to be filled in as we loop over times
        output = np.zeros((t.size, z0.size, zt.size))
        # Determine the transition rate matrix
        q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)
        for idx in range(t.size):
            # The transition probabilities follow from uniformization
            p_mat = p_mat_bld_uniform(q_mat, t[idx], k, num_states)
            # Fill in the output according to requested initial and final states (although many more
            # probabilities are actually computed)
            output[idx, :, :] = p_mat[np.ix_(np.array(z0 - z_min, dtype=np.int32),
                                             np.array(zt - z_min, dtype=np.int32))]
    return output



import numpy as np
import birdepy.utility as ut
from birdepy.interface_aug import f_fun_bld as f_fun_bld


def discrete_est_em_qn1(sorted_data, p0, likelihood, technique, known_p,
                        idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds,
                        con, max_it, i_tol, j_tol, h_tol, display, opt_method,
                        options):
    """
    Executes an expectation-maximization algorithm, which is accelerated using
    conjugate gradient ideas, to estimate parameters for a
    population-size-dependent birth-and-death process.

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.

    [1] Jamshidian, M., & Jennrich, R. I. (1997). Acceleration of the EM algorithm
     by using quasi‐Newton methods. Journal of the Royal Statistical Society:
     Series B (Statistical Methodology), 59(3), 569-587.

    """
    #Initialize a list to return parameters in each iteration and store the initial parameters in it
    iterations = [p0]

    it = 1
    # Set the difference to ensure at least one iteration occurs
    difference = i_tol + 1
    # Initilize estimate
    p_est = p0

    # Build surrogate log-likelihood function (defined in Section 2.1 of ref [1])
    f_fun = f_fun_bld(sorted_data, p_est, b_rate, d_rate, likelihood,
                      technique, idx_known_p, known_p, model, z_trunc,
                      j_tol, h_tol, options)[0]

    # Negate the surrogate log-likelihood function so that scipy.optimize.minimize()
    # Can be used to find a maximum
    def error_fun(p_prop):
        return -f_fun(p_prop)

    # Find a zero of an EM step (defined as $\tilde g(\theta) at start of Section 2.2
    # of reference [1])
    g_tilde = ut.minimize_(error_fun, p_est, p_bounds, con, opt_method,
                           options).x - p_est

    A = -np.identity(p0.size)

    while difference > i_tol and it < max_it:

        p_est_change_0 = np.matmul(-A, g_tilde)

        # First part of step (a) of QN1 algorithm in reference [1]
        # Ensure the next parameter estimate satisfies bounds and constraints
        m = 0
        feasible = False
        while not feasible:
            m += 1
            p_est_change = np.multiply(0.5 ** m, p_est_change_0)
            p_est_next = p_est + p_est_change
            feasible = True
            if type(con) == dict:
                if con['fun'](p_est_next) < 0:
                    feasible = False
            else:
                for c in con:
                    if c['fun'](p_est_next) < 0:
                        feasible = False
                        break
            if feasible:
                for idx, b in enumerate(p_bounds):
                    if p_est_next[idx] < b[0]:
                        feasible = False
                        break
                    if p_est_next[idx] > b[1]:
                        feasible = False
                        break

        # Negate the surrogate log-likelihood function so that scipy.optimize.minimize()
        # Can be used to find a maximum
        f_fun = f_fun_bld(sorted_data, p_est_next, b_rate, d_rate,
                          likelihood, technique, idx_known_p, known_p, model,
                          z_trunc, j_tol, h_tol, options)[0]

        # Negate the surrogate log-likelihood function so that scipy.optimize.minimize()
        # Can be used to find a maximum
        def error_fun(p_prop):
            return -f_fun(p_prop)

        # Second part of step (a) of QN1 algorithm in reference [1]
        g_tilde_change = ut.minimize_(error_fun, p_est_next, p_bounds,
                                      con, opt_method, options).x \
                         - p_est_next - g_tilde

        # Step (b) of QN1 algorithm in reference [1]
        A = A + np.outer((p_est_change - np.matmul(A, g_tilde_change)),
                          np.matmul(p_est_change, A)) / \
            np.inner(np.matmul(p_est_change, A), g_tilde_change)

        g_tilde += g_tilde_change

        # Compute the difference between this iteration estimate and
        # previous iteration estimate
        difference = np.sum(np.abs(p_est_next - p_est))
        # Prepare for next iteration
        p_est = p_est_next
        # Store the estimate
        iterations.append(p_est)

        # Print a progress update if requested
        if display:
            print('Iteration ', it, ' estimate is: ', p_est)
        it += 1

    return list(p_est), np.array(iterations)



import numpy as np
import birdepy.utility as ut
from birdepy.interface_aug import f_fun_bld as f_fun_bld
from birdepy.interface_dnm import bld_ll_fun
import birdepy.utility_em as utem


def discrete_est_em_Lange(data, sorted_data, p0, likelihood, technique,
                          known_p, idx_known_p, model, b_rate, d_rate, z_trunc,
                          p_bounds, con, max_it, i_tol, j_tol, h_tol, display,
                          opt_method, options):
    """
    Executes an expectation-maximization algorithm, which is accelerated using
    conjugate gradient ideas, to estimate parameters for a
    population-size-dependent birth-and-death process.

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.

    [1] Lange, K. (1995). A quasi-Newton acceleration of the EM algorithm.
     Statistica sinica, 1-18.
    """
    #Initialize a list to return parameters in each iteration and store the initial parameters in it
    iterations = [p0]

    it = 1
    p_est_it_m2 = p0

    # Build the log-likelihood function assuming all parameters are known
    pre_ll = bld_ll_fun(data, likelihood, model, z_trunc, options)

    # Adjust the log-likelihood function to account for unknown parameters and
    # negate it so that the minimize function performs maximisation.
    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll(param)

    # Build surrogate log-likelihood at initial (defined in Section 2 of ref [1]).
    f_fun_it_m2 = f_fun_bld(sorted_data, p_est_it_m2, b_rate, d_rate,
                            likelihood, technique, idx_known_p, known_p, model,
                            z_trunc, j_tol, h_tol, options)[0]
    # Optimize surrogate log-likelihood to find an estimate.
    p_est_it_m1 = ut.minimize_(lambda p: -f_fun_it_m2(p), p0, p_bounds, con,
                               opt_method, options).x

    # Print a progress update if requested
    if display:
        print('Iteration ', it, ' estimate is: ', p_est_it_m1)
    it += 1

    # Initial approximation to the H quantity in Equation (3) of reference [1].
    # This choice is discussed after Equation (8) in reference [1].
    B = np.zeros((p0.size, p0.size))

    # Compute difference to determine if should terminate.
    diff_one = np.sum(np.abs(p_est_it_m1 - p_est_it_m2))
    # This is s as defined by Equation (9) in reference [1].
    s = p_est_it_m1 - p_est_it_m2

    while diff_one > i_tol and it < max_it:
        # Surrogate log-likelihood
        f_fun_it_m1 = f_fun_bld(sorted_data, p_est_it_m1, b_rate, d_rate,
                                likelihood, technique, idx_known_p, known_p,
                                model, z_trunc, j_tol, h_tol, options)[0]

        # Hessian of surrogate log-likelihood (Equation (5) in reference [1]).
        H_f = ut.Hessian(f_fun_it_m1, p_est_it_m1, p_bounds)

        # This is g as defined by Equation (9) in reference [1].
        g = ut.Jacobian(f_fun_it_m1, p_est_it_m2, p_bounds) - \
              ut.Jacobian(f_fun_it_m2, p_est_it_m2, p_bounds)

        # This is c and v as defined by Equation (11) in reference [1].
        c = 1 / np.inner(g - np.matmul(B, s), s)
        v = g - np.matmul(B, s)

        # Update approximation to the H quantity in Equation (3) of reference [1].
        # This is Equation (10) in reference [1].
        B = B + c * np.outer(v, v)

        # Modify approximation to ensure it is negative definite (as discussed
        # at the end of Section 3 of reference [1])
        m = 1
        H_f_mod = H_f - np.multiply(0.5 ** m, B)
        while not np.all(np.linalg.eigvals(H_f_mod) < 0):
            m += 1
            H_f_mod = H_f - np.multiply(0.5 ** m, B)

        # Jacobian of log-likelihood helps determine direction
        J_ell = ut.Jacobian(ll, p_est_it_m1, p_bounds)

        # Determine direction to change parameters (Equation (12) of reference [1]).
        direction = -np.matmul(np.linalg.inv(H_f_mod), J_ell)

        # Need to store this since each iteration depends on previous 2 iterations
        p_est_it_m2 = p_est_it_m1

        # Tells us how much to change parameters by (Equation (4) in reference [2])
        a_opt = utem.line_search_ll(direction, data, p_est_it_m1, likelihood,
                                    known_p, idx_known_p, model, z_trunc,
                                    p_bounds, con, opt_method, options)[0]

        p_est_it_m1 = p_est_it_m1 + np.multiply(a_opt, direction)

        # Ensure estimate satisfied bounds
        for idx in range(p_est_it_m1.size):
            p_est_it_m1[idx] = min(p_bounds[idx][1],
                                   max(p_bounds[idx][0],
                                       p_est_it_m1[idx]))

        # Store estimate
        iterations.append(p_est_it_m1)

        # Compute difference to determine if should terminate
        diff_one = np.sum(np.abs(p_est_it_m1 - p_est_it_m2))
        # This is s as defined by Equation (9) in reference [1]
        s = p_est_it_m1 - p_est_it_m2
        if display:
            print('Iteration ', it, ' estimate is: ', p_est_it_m1)
        it += 1

    return list(p_est_it_m1), np.array(iterations)



import numpy as np
import birdepy.em_none as em_none
import birdepy.em_Lange as em_Lange
import birdepy.em_qn1 as em_qn1
import birdepy.em_qn2 as em_qn2
import birdepy.em_cg as em_cg
import birdepy.utility as ut
from birdepy.interface_dnm import bld_ll_fun


def discrete_est_em(data, p0, technique, accelerator, likelihood,
                    p_bounds, con, known_p, idx_known_p, model, b_rate, d_rate,
                    z_trunc, max_it, i_tol, j_tol, h_tol, display, opt_method,
                    options):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using the expectation maximization algorithm.
    See :ref:`here <Expectation Maximization>` for more information.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'em'::

        bd.estimate(t_data, p_data, p0, p_bounds, framework='em', technique='expm',
                         accelerator='none', likelihood='expm', laplace_method='cme-talbot',
                         lentz_eps=1e-6, max_it=25, i_tol=1e-2, j_tol=1e-1, h_tol=1e-2,
                         z_trunc=())

    The parameters associated with this framework (listed below) can be
    accessed using kwargs in :func:`birdepy.estimate()`. See documentation of
    :func:`birdepy.estimate` for the main arguments.

    Parameters
    ----------
    technique : string, optional
        Expectation step technique. Should be one of
        (alphabetical ordering):

            - 'expm' (default)
            - 'ilt'
            - 'num'

        See :ref:`here <Expectation Step Techniques>` for more information.

    accelerator : string, optional
        EM accelerator method. Should be one of
        (alphabetical ordering):

            - 'cg' (see [3])
            - 'none' (see [4])
            - 'Lange' (default) (see [5])
            - 'qn1' (see [6])
            - 'qn2' (see [6])

        See :ref:`here <Acceleration>` for more information.

    likelihood : string, optional
        Likelihood approximation method. Should be one of
        (alphabetical ordering):

            - 'da' (default) (see :ref:`here <birdepy.probability(method='da')>`)
            - 'Erlang' (default) (see :ref:`here <birdepy.probability(method='Erlang')>`)
            - 'expm' (see :ref:`here <birdepy.probability(method='expm')>`)
            - 'gwa' (see :ref:`here <birdepy.probability(method='gwa')>`)
            - 'gwasa' (see :ref:`here <birdepy.probability(method='gwasa')>`)
            - 'ilt' (see :ref:`here <birdepy.probability(method='ilt')>`)
            - 'oua' (see :ref:`here <birdepy.probability(method='oua')>`)
            - 'uniform' (see :ref:`here <birdepy.probability(method='uniform')>`)

        The links point to the documentation of the relevant `method` in
        :func:`birdepy.probability`. The arguments associated with each of
        these methods may be used as a kwarg in :func:`birdepy.estimate()`
        when `likelihood` is set to use the method.

    laplace_method : string, optional
        Numerical inverse Laplace transform algorithm to use. Should be one of:
        'cme-talbot' (default), 'cme', 'euler', 'gaver', 'talbot', 'stehfest',
        'dehoog', 'cme-mp' or 'gwr'.

    lentz_eps : scalar, optional
        Termination threshold for Lentz algorithm computation of Laplace
        domain functions.

    max_it : int, optional
        Maximum number of iterations of the algorithm to perform.

    i_tol : scalar, optional
        Algorithm terminates when ``sum(abs(p(i) - p(i-1)) < i_tol``
        where `p(i)` and `p(i-1)` are estimates corresponding to iteration `i`
        and `i-1`.

    j_tol : scalar, optional
        States with expected number of upward transitions or expected number
        of downward transitions greater than `j_tol` are included in E steps.

    h_tol : scalar, optional
        States with expected holding time greater than `h_tol` are included in
        E steps.

    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default
        ``z_trunc=[z_min, z_max]`` where
        ``z_min=max(0, min(p_data) - 2*z_range)``
        and ``z_max=max(p_data) + 2*z_range`` where
        ``z_range=max(p_data)-min(p_data)``.

    Examples
    --------
    Use :func:`birdepy.simulate.discrete` to simulate a sample path of the 'Verhulst 2 (SIS)'
    model: ::

        import birdepy as bd
        t_data = list(range(100))
        p_data = bd.simulate.discrete([0.75, 0.25, 0.02, 1], 'Ricker', 10, t_data,
                                      survival=True, seed=2021)

    Estimate the parameter values from the simulated data using the 'em'
    `framework` with various `technique` and `accelerator` approaches: ::

        for technique in ['expm', 'ilt', 'num']:
            for accelerator in ['cg', 'none', 'Lange', 'qn1', 'qn2']:
                tic = time.time()
                est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[1e-6,1], [1e-6,1], [1e-6, 0.1]],
                                  framework='em', technique=technique, accelerator=accelerator,
                                  model='Ricker', idx_known_p=[3], known_p=[1])

    The constraint ``con={'type': 'ineq', 'fun': lambda p: p[0]-p[1]}`` ensures
    that p[0] > p[1] (i.e., rate of spread greater than recovery rate).

    A ``RuntimeWarning`` associated with SciPy's :func:`minimize` function  may
    appear, this can be ignored.

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Jamshidian, M. and  Jennrich, R.I. Conjugate gradient acceleration
     of the EM algorithm. Journal of the American Statistical Association,
     88(421):221-228, 1993.

    .. [4] Dempster, A.P., Laird, N.M. and Rubin, D.B. Maximum likelihood from
     incomplete data via the EM algorithm. Journal of the Royal Statistical
     Society: Series B (Methodological), 39(1):1-22, 1977.

    .. [5] Lange, K. A quasi-Newton acceleration of the EM algorithm.
     Statistica Sinica, 1-18, 1995.

    .. [6] Jamshidian, M. and  Jennrich, R.I. Acceleration of the EM algorithm
      by using quasi-Newton methods. Journal of the Royal Statistical
      Society: Series B (Methodological), 59(3):569-587, 1997.

    """
    # Sort the data into a format suitable for EM algorithms
    sorted_data = ut.data_sort_2(data)
    # Each acceleration method is coded independently and stored in its own module.
    # Inside this control flow parameter estimates are determined and the values of estimates in
    # all iterations of the EM algorithm are also returned
    if accelerator == 'none':
        p_est, iterations = em_none.discrete_est_em_none(
            sorted_data, p0, likelihood, technique, known_p, idx_known_p,
            model, b_rate, d_rate, z_trunc, p_bounds, con, max_it, i_tol,
            j_tol, h_tol, display, opt_method, options)

    elif accelerator == 'cg':
        p_est, iterations = em_cg.discrete_est_em_cg(
            data, sorted_data, p0, likelihood, technique, known_p,
            idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds,
            con, max_it, i_tol, j_tol, h_tol, display, opt_method, options)

    elif accelerator == 'Lange':
        p_est, iterations = em_Lange.discrete_est_em_Lange(
            data, sorted_data, p0, likelihood, technique, known_p,
            idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds, con,
            max_it, i_tol, j_tol, h_tol, display, opt_method, options)

    elif accelerator == 'qn1':
        p_est, iterations = em_qn1.discrete_est_em_qn1(
            sorted_data, p0, likelihood, technique, known_p,
            idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds,
            con, max_it, i_tol, j_tol, h_tol, display, opt_method, options)

    elif accelerator == 'qn2':
        p_est, iterations = em_qn2.discrete_est_em_qn2(
            data, sorted_data, p0, likelihood, technique, known_p,
            idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds, con,
            max_it, i_tol, j_tol, h_tol, display, opt_method, options)

    else:
        raise ValueError("Argument 'accelerator' has an unknown value. Should "
                         "be one of: 'none', 'cg', 'Lange', 'qn1' or 'qn2'.")

    # This part of the code is for finding the covariance matrix and log-likelihood at the estimate
    pre_ll_fun = bld_ll_fun(data, likelihood, model, z_trunc, options)

    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll_fun(param)

    for idx in range(len(p0)):
        p_est[idx] = min(p_bounds[idx][1],
                         max(p_bounds[idx][0], p_est[idx]))

    if p0.size > 1:
        try:
            cov = np.linalg.inv(-ut.Hessian(ll, p_est, p_bounds))
        except:
            cov = 'Covariance matrix could not be determined.'
    else:
        try:
            print(np.array([-1/ut.Hessian(ll, p_est, p_bounds)]))
            cov = np.array([-1/ut.Hessian(ll, p_est, p_bounds)])
        except:
            cov = 'Covariance matrix could not be determined.'

    ll = ll(np.array(p_est))

    return p_est, cov, ll, iterations



import numpy as np
import mpmath as mp
from scipy.optimize import minimize, differential_evolution, NonlinearConstraint
from scipy.optimize._numdiff import approx_derivative
from scipy.stats import multivariate_normal
import math
from birdepy.iltcme import IltCmeParams
from collections import Counter
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
from matplotlib.patches import Ellipse
from gwr_inversion import gwr


def Jacobian(fun, x, bounds):
    """
    Approximates the Jacobian of 'fun' at point 'x' without
    evaluating the function outside of 'bounds'.
    """
    bounds = [[b[0] for b in bounds], [b[1] for b in bounds]]
    return approx_derivative(fun, x, bounds=bounds)


def Hessian(fun, x, bounds):
    """
    Approximates the Hessian of 'fun' at point 'x' without
    evaluating the function outside of 'bounds'.
    """
    bounds = [[b[0] for b in bounds], [b[1] for b in bounds]]
    return approx_derivative(lambda y: approx_derivative(fun,
                                                         y,
                                                         bounds=bounds),
                             x, bounds=bounds)


def confidence_region(mean, cov, obs, se_type, xlabel, ylabel, export):
    """
    Determines and plots confidence regions for a multivariate normal
    distribution with expected value 'mean' and covariance 'cov'.
    """
    [Eval, Evec] = np.linalg.eig(cov)
    Evec = -np.hstack((Evec[:, 1].reshape((2, 1)), Evec[:, 0].reshape((2, 1))))
    xCenter = mean[0]
    yCenter = mean[1]
    theta = np.linspace(0, 2 * np.pi, 100)
    x_vec = np.array([1, 0])
    cosrotation = np.dot(x_vec, Evec[:, 1]) / \
                  (np.linalg.norm(x_vec) * np.linalg.norm(Evec[:, 1]))
    rotation = np.pi / 2 - np.arccos(cosrotation)

    R = np.array([[np.sin(rotation), np.cos(rotation)],
                  [-np.cos(rotation), np.sin(rotation)]])

    chisq = [1.368, 3.2188, 5.991]

    x = np.empty((len(theta), len(chisq)))
    y = np.empty((len(theta), len(chisq)))
    xRadius = np.empty(len(chisq))
    yRadius = np.empty(len(chisq))

    x_plot = np.empty((len(theta), len(chisq)))
    y_plot = np.empty((len(theta), len(chisq)))

    rotated_Coords = np.empty((2, len(theta), len(chisq)))

    for i in range(len(chisq)):
        xRadius[i] = np.sqrt(chisq[i] * Eval[0])
        yRadius[i] = np.sqrt(chisq[i] * Eval[1])

        x[:, i] = xRadius[i] * np.cos(theta)
        y[:, i] = yRadius[i] * np.sin(theta)

        rotated_Coords[:, :, i] = np.matmul(R, np.row_stack((x[:, i], y[:, i])))

        x_plot[:, i] = rotated_Coords[0, :, i].T + xCenter
        y_plot[:, i] = rotated_Coords[1, :, i].T + yCenter

    fig, (ax) = plt.subplots(figsize=(7, 5))

    labels = ['$50\%$', '$80\%$', '$95\%$']
    linestyles = [':', '--', '-']

    for i in range(len(chisq)):
        ax.plot(x_plot[:, i], y_plot[:, i], label=labels[i],
                linestyle=linestyles[i], color='k')
    ax.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.plot(mean[0], mean[1], '+k', label="Estimate")
    if obs is not None:
        ax.scatter(obs[:, 0], obs[:, 1], c="b", marker="^", s=5,
                   label="Simulated\n samples")
    if se_type == "asymptotic":
        ax.set_title("Asymptotic confidence region")
    elif se_type == "simulated":
        ax.set_title("Simulated confidence region")
    ax.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
    fig.set_tight_layout({
        'pad': 0
    })
    if isinstance(export, str):
        import tikzplotlib
        tikzplotlib.save(export + ".tex")
    return


def data_sort(t_data, p_data):
    """
    Data sorter to improve the efficiency of the :func:`birdepy` package.
    Counts how many times each probability_ z0 -> zt occurs, where transitions
    occurring over different elapsed times are treated as non-equal.

    Parameters
    ----------
    t_data : array_like
        Observation times of birth-and-death process.
        If one trajectory is observed, then this is a list.
        If multiple trajectories are observed, then this is a list of lists
        where each list corresponds to a trajectory.

    p_data : array_like
        Observed populations of birth-and-death process at times in argument
        't_data'.
        If one trajectory is observed, then this is a list.
        If multiple trajectories are observed, then this is a list of lists
        where each list corresponds to a traj  ectory.

    Return
    -------
    data : collections.Counter object
        A dictionary with keys corresponding to observed transitions
        (z0, zt, t) and values corresponding to the number of times
        an observed transition occurs.

    """
    if type(t_data) is np.ndarray:
        t_data = t_data.tolist()
    if type(t_data[0]) is not list:
        t_data = [t_data]
    if type(p_data) == np.ndarray:
        p_data = p_data.tolist()
    if type(p_data[0]) is not list:
        p_data = [p_data]

    all_transitions = []
    for idx1 in range(len(p_data)):
        for idx2 in range(len(p_data[idx1]) - 1):
            all_transitions.append((p_data[idx1][idx2],
                                    p_data[idx1][idx2 + 1],
                                    t_data[idx1][idx2 + 1] -
                                    t_data[idx1][idx2]))
    return Counter(all_transitions)


def data_sort_2(data):
    """
    Further sorts output of :func:`birdepy._utility.data_sort`. Separates data
    into sub-collections where all transitions in any given sub-collection
    occur in the same elapsed time (rounded to 8 decimal places). This is useful
    when computing transition probabilities using methods that compute all
    transition probabilities simultaneously for a specified time (e.g., the
    matrix exponential method).

    Parameters
    ----------
    data : collections.Counter object
        A dictionary with keys corresponding to observed transitions
        (z0, zt, t) and values corresponding to the number of times
        a probability_ is observed.

    Return
    -------
    sorted_data : dict
        A dictionary with keys corresponding to inter-observation times and
        values containing dict objects with keys corresponding to observed
        (z0, zt) pairs and values corresponding to the number of times a
        pair is observed for the relevant inter-observation time. E.g.,
        {t0 : {(a, b): 3, (c, d) : 4}, t1 : {(e, f) : 1, (g, h) : 10}}.

    """
    sorted_data = {}
    for i in data:
        if np.around(i[2], 8) not in sorted_data:
            sorted_data[np.around(i[2], 8)] = {(i[0], i[1]): data[i]}
        else:
            sorted_data[np.around(i[2], 8)][(i[0], i[1])] = data[i]
    return sorted_data


def higher_birth(model):
    """
    Returns a function that gives the birth rate as a function
    of population size and parameters.
    """
    if model == 'Verhulst':
        return lambda z, p: ((p[2]*z) <= 1) * p[0] * (1 - p[2]*z) * z
    elif model == 'Ricker':
        return lambda z, p: p[0] * np.exp(-(p[2] * z)**p[3]) * z
    elif model == 'Hassell':
        return lambda z, p: \
            p[0] * z / (1 + p[2] * z) ** p[3]
    elif model == 'MS-S':
        return lambda z, p: \
            p[0] * z / (1 + (p[2] * z) ** p[3])
    elif model == 'Moran':
        return lambda z, p: (z <= p[4]) * (
            (p[4] - z) * (p[0] * z * (1 - p[2]) / p[4] + p[1] * (p[4] - z) * p[3] / p[4]) / p[4])
    elif model == 'pure-birth':
        return lambda z, p: p[0] * z
    elif model == 'pure-death':
        return lambda z, p: 0
    elif model == 'Poisson':
        return lambda z, p: p[0]
    elif model == 'linear':
        return lambda z, p: p[0] * z
    elif model == 'linear-migration':
        return lambda z, p: p[0] * z + p[2]
    elif model == 'M/M/1':
        return lambda z, p: p[0]
    elif model == 'M/M/inf':
        return lambda z, p: p[0]
    elif model == 'loss-system':
        return lambda z, p: (z < p[2]) * p[0]
    else:
        raise TypeError("Argument 'model' has an unknown value.")


def higher_death(model):
    """
    Returns a function that gives the death rate as a function
    of population size and parameters.
    """
    if model == 'Verhulst':
        return lambda z, p: p[1] * (1 + p[3]*z) * z
    elif model in ['Ricker', 'Hassell',
                   'MS-S', 'loss-system', 'linear',
                   'linear-migration', 'M/M/inf']:
        return lambda z, p: p[1] * z
    elif model == 'Moran':
        return lambda z, p: (z <= p[4]) * (
            z * (p[2] * (p[4] - z) * (1 - p[3]) / p[4] +
                 p[0] * z * p[2] / p[4]) / p[4])
    elif model == 'M/M/1':
        return lambda z, p: (z > 0) * p[1]
    elif model in ['pure-birth', 'Poisson']:
        return lambda z, p: 0
    elif model == 'pure-death':
        return lambda z, p: p[0] * z
    else:
        raise TypeError("Argument 'model' has an unknown value.")


def higher_h_fun(model):
    """
    Returns a function that gives the derivative with respect
    to population size of the function given by the birth rate
    minus the death rate.
    """
    if model == 'Verhulst':
        return lambda z, p: p[0]-p[1]-2*z*(p[0]*p[2]+p[1]*p[3])
    elif model == 'Ricker':
        return lambda z, p: p[0] * np.exp(-(p[2] * z)**p[3]) * (1 - p[3]*(p[2] * z)**p[3]) - p[1]
    elif model == 'Hassell':
        return lambda z, p: (p[0] * (1 + p[2]*z - p[3] * z)) / \
                            (1 + (p[2] * z)) ** p[3] - p[1]
    elif model == 'MS-S':
        return lambda z, p: (p[0] * (1 + (p[2] * z) ** p[3] * (1 - p[3]))) / \
                            (1 + (p[2] * z) ** p[3]) ** 2
    elif model == 'Moran':
        return lambda z, p: (z*p[0]*(p[4]*(1-p[2])-z) + p[1]*(z-[4])*(z-p[4]*p[3]))\
                            / p[4]**2
    elif model == 'linear':
        return lambda z, p: (p[0] - p[1]) * (z - z + 1)
    elif model == 'linear-migration':
        return lambda z, p: (p[0] - p[1]) * (z - z + 1)
    elif model == 'M/M/1':
        return lambda z, p: 0 * (z - z + 1)
    elif model == 'M/M/inf':
        return lambda z, p: -p[1] * (z - z + 1)
    elif model == 'pure-birth':
        return lambda z, p: p[0] * (z - z + 1)
    elif model == 'pure-death':
        return lambda z, p: -p[0] * (z - z + 1)
    elif model == 'Poisson':
        return lambda z, p: 0 * (z - z + 1)
    else:
        raise TypeError("Unknown model.")


def higher_zf_bld(model):
    """
    Returns the fixed points of the fluid model associated with the model.
    """
    if model == 'Verhulst':
        return lambda p: [0, (p[0]-p[1])/(p[2]*p[0] + p[1]*p[3])]
    elif model == 'Ricker':
        return lambda p: [0, ((np.log(p[0] / p[1]))**(1/p[3])) / p[2]]
    elif model == 'Hassell':
        return lambda p: [0, ((p[0] / p[1]) ** (1 / p[3]) - 1)/p[2]]
    elif model == 'MS-S':
        return lambda p: [0, ((p[0] / p[1] - 1) ** (1 / p[3]))/p[2]]
    elif model == 'Moran':
        return lambda p: [p[4]*(p[0]*(1-p[2])-p[1]*(1+p[3]) + np.sqrt((p[2]-1)**2 * p[0]**2 + 2 * p[0] * p[1]*(p[2]+p[3]+p[2]*p[3]-1) + p[1] ** 2 *(p[3]-1)**2))/(2*(p[0]-p[1])),
                          p[4]*(p[0]*(1-p[2])-p[1]*(1+p[3]) - np.sqrt((p[2]-1)**2 * p[0]**2 + 2 * p[0] * p[1]*(p[2]+p[3]+p[2]*p[3]-1) + p[1] ** 2 *(p[3]-1)**2))/(2*(p[0]-p[1]))]
    elif model == 'linear':
        return lambda p: [0]
    elif model == 'linear-migration':
        return lambda p: [-p[2] / (p[0] - p[1])]
    elif model == 'Poisson':
        return lambda p: [np.nan]
    elif model == 'M/M/1':
        return lambda p: [np.nan]
    elif model == 'M/M/inf':
        return lambda p: [p[0] / p[1]]
    elif model == 'pure-birth':
        return lambda p: [np.nan]
    elif model == 'pure-death':
        return lambda p: [np.nan]
    elif model == 'loss-system':
        return lambda p: [np.nan]
    else:
        raise TypeError("Unknown model.")


def q_mat_bld(z_min, z_max, p, b_rate, d_rate):
    """
    Builds the transition rate matrix for a
    population-size-dependent birth-and-death process.
    """
    num_states = int(z_max - z_min + 1)
    states = np.arange(z_min, z_max + 1, 1)
    q_mat = np.zeros((num_states, num_states))
    q_mat[-1, -1] = -d_rate(z_max, p)
    q_mat[-1, -2] = d_rate(z_max, p)
    q_mat[0, 0] = -b_rate(z_min, p)
    q_mat[0, 1] = b_rate(z_min, p)
    for z in np.arange(1, num_states - 1, 1):
        _state = states[z]
        q_mat[z, z] = -b_rate(_state, p) - d_rate(_state, p)
        q_mat[z, z + 1] = b_rate(_state, p)
        q_mat[z, z - 1] = d_rate(_state, p)
    return q_mat


def p_bld(p_prop, idx_known_p, known_p):
    """
    Augments parameters 'p_prop' with parameters 'known_p'
    to form a complete vector of parameters.
    """
    known_p = np.array(known_p)
    p_size = known_p.size + p_prop.size
    idx_known_p = list(idx_known_p)
    if len(known_p) != len(idx_known_p):
        raise TypeError("Argument 'idx_known_p' and argument 'known_p' must"
                        " be the same size.")
    idx_unknown_p = list(set(range(p_size)) - set(idx_known_p))
    p = np.empty(p_size)
    p[idx_known_p] = known_p
    p[idx_unknown_p] = p_prop
    return p


def laplace_invert_mexp(fun, t, max_fun_evals, method="cme"):
    """
    Numerical Laplace transform inversion using the method of reference [1],
    an Euler scheme or a Gaver scheme.

    Uses code downloaded from https://inverselaplace.org/ in June of 2021.

    [1] Horváth, G., Horváth, I., Almousa, S. A. D., & Telek, M. (2020).
    Numerical inverse Laplace transformation using concentrated matrix
    exponential distributions. Performance Evaluation, 137, 102067.
    """
    if method == "cme":
        # find the most steep CME satisfying max_fun_evals
        params = IltCmeParams.params[0]
        for p in IltCmeParams.params:
            if p["cv2"] < params["cv2"] and (p["n"] + 1) <= max_fun_evals:
                params = p
        eta = np.concatenate(([params["c"]],
                              np.array(params["a"]) + 1j * np.array(
                                  params["b"]))) * params["mu1"]
        # print('eta:', eta)
        beta = np.concatenate(
            ([1], 1 + 1j * np.arange(1, params["n"] + 1)
             * params["omega"])) * params["mu1"]
        # print('beta:', beta)
    elif method == "euler":
        n_euler = math.floor((max_fun_evals - 1) / 2)
        eta = np.concatenate(
            ([0.5], np.ones(n_euler), np.zeros(n_euler - 1), [2 ** -n_euler]))
        logsum = np.cumsum(np.log(np.arange(1, n_euler + 1)))
        for k in range(1, n_euler):
            eta[2 * n_euler - k] = eta[2 * n_euler - k + 1] + math.exp(
                logsum[n_euler - 1] - n_euler * math.log(2.0) - logsum[k - 1] -
                logsum[n_euler - k - 1])
        k = np.arange(2 * n_euler + 1)
        beta = n_euler * math.log(10.0) / 3.0 + 1j * math.pi * k
        eta = (10 ** ((n_euler) / 3.0)) * (1 - (k % 2) * 2) * eta
    elif method == "gaver":
        if max_fun_evals % 2 == 1:
            max_fun_evals -= 1
        ndiv2 = int(max_fun_evals / 2)
        eta = np.zeros(max_fun_evals)
        beta = np.zeros(max_fun_evals)
        logsum = np.concatenate(
            ([0], np.cumsum(np.log(np.arange(1, max_fun_evals + 1)))))
        for k in range(1, max_fun_evals + 1):
            insisum = 0.0
            for j in range(math.floor((k + 1) / 2), min(k, ndiv2) + 1):
                insisum += math.exp(
                    (ndiv2 + 1) * math.log(j) - logsum[ndiv2 - j] + logsum[
                        2 * j] - 2 * logsum[j] - logsum[k - j] - logsum[
                        2 * j - k])
            eta[k - 1] = math.log(2.0) * (-1) ** (k + ndiv2) * insisum
            beta[k - 1] = k * math.log(2.0)
    else:
        raise TypeError('Unknown Laplace inversion method.')
    res = []
    for x in t:
        res.append(eta.dot([fun(b / x) for b in beta]).real / x)
    return res


def cme(fun, t, k):
    """
    Numerical Laplace transform inversion using the method of reference [1].

    Is based on code downloaded from https://inverselaplace.org/ in June of 2021.

    [1] Horváth, G., Horváth, I., Almousa, S. A. D., & Telek, M. (2020).
    Numerical inverse Laplace transformation using concentrated matrix
    exponential distributions. Performance Evaluation, 137, 102067.
    """
    if 'cme_params' not in globals() or globals()['cme_k'] != k:
        global cme_params
        global cme_k
        global cme_eta
        global cme_beta
        cme_k = k
        cme_params = IltCmeParams.params[0]
        for p in IltCmeParams.params:
            if p['cv2'] < cme_params['cv2'] and (p['n'] + 1) <= cme_k:
                cme_params = p
        cme_eta = [cme_params['a'][idx] + 1j * cme_params['b'][idx] for idx in
                   range(cme_params['n'])]
        cme_eta.insert(0, cme_params['c'] + 1j * 0)
        cme_eta = mp.matrix(cme_eta).T * cme_params['mu1']
        cme_beta = [1 + 1j * idx * cme_params['omega'] for idx in
                    range(1, cme_params['n'] + 1)]
        cme_beta.insert(0, 1)
        cme_beta = mp.matrix(cme_beta) * cme_params['mu1']
    foo = mp.matrix([fun(mp.fdiv(b, t)) for b in cme_beta])
    foo = cme_eta * foo
    return float(mp.fdiv(mp.re(foo[0]), t))


def laplace_invert(fun, t, **options):
    """
    Numerical Laplace transform inversion using a variety of different methods.
    """
    if 'laplace_method' in options.keys():
        laplace_method = options['laplace_method']
    else:
        laplace_method = 'cme-mp'
    if 'k' in options.keys():
        k = options['k']
    else:
        k = 25
    if laplace_method == 'gwr':
        return float(gwr(fun, time=t, M=k))
    elif laplace_method in ['cme', 'euler', 'gaver']:
        return laplace_invert_mexp(fun, [t], k, method=laplace_method)[0]
    elif laplace_method in ['talbot', 'stehfest', 'dehoog']:
        return float(mp.invertlaplace(fun, t, method=laplace_method))
    elif laplace_method == 'cme-mp':
        return cme(fun, t, k)
    elif laplace_method == 'cme-talbot':
        if 'f_bounds' in options.keys():
            f_min = options['f_bounds'][0]
            f_max = options['f_bounds'][1]
        else:
            raise TypeError("When 'laplace_method' is 'cme-talbot', kwarg"
                            "'f_bounds' must be given as a list [f_min, f_max]."
                            "When method 'cme-mp' returns a value outside "
                            "these bounds, method 'talbot' is used instead. ")
        output = cme(fun, t, k)
        if f_min <= output <= f_max:
            return output
        else:
            return float(mp.invertlaplace(fun, t, method='talbot'))
    else:
        raise TypeError('Unknown Laplace inversion method.')


def add_options(options):
    """
    Completes the 'options' dictionary for use in :func:`minimize_`.
    Is needed to enable BirDePy to specify default values for options
    in the SciPy optimization functions.
    """
    if 'seed' not in options.keys():
        options['seed'] = np.random.default_rng()
    if 'strategy' not in options.keys():
        options['strategy'] = 'best1bin'
    if 'maxiter' not in options.keys():
        options['maxiter'] = 1000
    if 'popsize' not in options.keys():
        options['popsize'] = 15
    if 'tol' not in options.keys():
        options['tol'] = 0.01
    if 'mutation' not in options.keys():
        options['mutation'] = 0.5
    if 'recombination' not in options.keys():
        options['recombination'] = 0.7
    if 'callback' not in options.keys():
        options['callback'] = None
    if 'disp' not in options.keys():
        options['disp'] = False
    if 'polish' not in options.keys():
        options['polish'] = True
    if 'init' not in options.keys():
        options['init'] = 'latinhypercube'
    if 'atol' not in options.keys():
        options['atol'] = 0
    if 'updating' not in options.keys():
        options['updating'] = 'immediate'
    if 'jac' not in options.keys():
        options['jac'] = '2-point'
    if 'maxcor' not in options.keys():
        options['maxcor'] = 10
    if 'ftol' not in options.keys():
        options['ftol'] = 2.220446049250313e-09
    if 'gtol' not in options.keys():
        options['gtol'] = 1e-05
    if 'eps' not in options.keys():
        options['eps'] = None
    if 'maxfun' not in options.keys():
        options['maxfun'] = 15000
    if 'maxiter' not in options.keys():
        options['maxiter'] = 15000
    if 'iprint' not in options.keys():
        options['iprint'] = -1
    if 'callback' not in options.keys():
        options['callback'] = None
    if 'maxls' not in options.keys():
        options['maxls'] = 20
    if 'finite_diff_rel_step' not in options.keys():
        options['finite_diff_rel_step'] = None
    if 'ftol' not in options.keys():
        options['ftol'] = 1e-06
    if 'eps' not in options.keys():
        options['eps'] = 1.4901161193847656e-08
    if 'disp' not in options.keys():
        options['disp'] = False
    if 'maxiter' not in options.keys():
        options['maxiter'] = 100
    if 'finite_diff_rel_step' not in options.keys():
        options['finite_diff_rel_step'] = None
    if 'jac' not in options.keys():
        options['jac'] = None
    return options


def minimize_(error_fun, p0, p_bounds, con, opt_method, options):
    """
    Minimizes 'error_fun' using the method specified by 'opt_method'.
    """
    if opt_method == 'differential-evolution':
        if con != ():
            old_con = con
            con= (NonlinearConstraint(con['fun'], 0, np.inf))
        sol = differential_evolution(error_fun, p_bounds, args=(),
                                     strategy=options['strategy'],
                                     maxiter=options['maxiter'],
                                     popsize=options['popsize'],
                                     tol=options['tol'],
                                     mutation=options['mutation'],
                                     recombination=options[
                                         'recombination'],
                                     seed=options['seed'],
                                     callback=options['callback'],
                                     disp=options['disp'],
                                     polish=False,
                                     init=options['init'],
                                     atol=options['atol'],
                                     updating=options['updating'],
                                     constraints=con)
        if options['polish']:
            if con != ():
                sol = minimize(error_fun, sol.x, method='SLSQP', bounds=p_bounds,
                               constraints=old_con)
            else:
                sol = minimize(error_fun, sol.x, method='L-BFGS-B', bounds=p_bounds)
    elif opt_method == 'L-BFGS-B':
        sol = minimize(error_fun, p0, method=opt_method, bounds=p_bounds,
                       callback=options['callback'],
                       jac=options['jac'],
                       options={'maxcor': options['maxcor'],
                                'ftol': options['ftol'],
                                'gtol': options['gtol'],
                                'eps': options['eps'],
                                'maxfun': options['maxfun'],
                                'maxiter': options['maxiter'],
                                'iprint': options['iprint'],
                                'maxls': options['maxls'],
                                'finite_diff_rel_step':
                                    options['finite_diff_rel_step']})
    elif opt_method == 'SLSQP':
        sol = minimize(error_fun, p0, method=opt_method, bounds=p_bounds,
                       constraints=con,
                       jac=options['jac'],
                       options={'ftol': options['ftol'],
                                'eps': options['eps'],
                                'disp': options['disp'],
                                'maxiter': options['maxiter'],
                                'finite_diff_rel_step':
                                    options['finite_diff_rel_step']})
    else:
        sol = minimize(error_fun, p0, method=opt_method, bounds=p_bounds,
                       constraints=con)
    return sol


def trap_int(y, x):
    """
    Numerical integration of sampled data using the trapezoidal rule.

    Parameters
    ----------
    y : array_like
        Function values.

    x : array_like
        Domain values.

    Return
    -------
    res : ndarray
        Cumulative integral of `y` at points in `x`.
    """
    x = np.array(x)
    y = np.array(y)
    res = np.hstack(([0], np.cumsum(np.multiply(0.5,
                                                np.multiply(np.diff(x),
                                                            np.add(y[1:],
                                                                   y[0:-1])))))
                    )
    return res



import numpy as np
from scipy.integrate import solve_ivp
import warnings


def sq_bld(data, b_rate, d_rate):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using the *fluid model least lse estimation*
    method.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'lse' and `squares` set to 'fm': ::

        bd.estimate(t_data, p_data, p0, p_bounds, framework='lse', squares='fm')

    This function does not have any arguments which are not already described
    by the documentation of ``birdepy.estimate`` (see :ref:`here <Estimation>`)
    or use ``help(birdepy.estimate)``.

    Examples
    --------

    See also
    --------
    birdepy.estimate
    birdepy.forecast

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    # These are methods used to solve differential equations as used by scipy.integrate.solve_ivp()
    solver_methods = ['RK45', 'Radau', 'RK23', 'BDF', 'DOP853']

    def error_fun(param):
        err = 0
        for i in data:
            for meth in solver_methods:
                fluid_path = solve_ivp(
                    lambda t, z: b_rate(z, param) - d_rate(z, param),
                    [0, i[2]+1e-100],
                    [i[0]],
                    t_eval=[i[2]],
                    method=meth)
                if fluid_path.success:
                    err += np.square(fluid_path.y[0][0] - i[1])
                    break
            if not fluid_path.success:
                err += 1
                warnings.warn(
                    "Failed to find a solution to an ordinary "
                    "differential equation, some output has been replaced"
                    " by a default value. ",
                    category=RuntimeWarning)
        return err

    return lambda param_prop: error_fun(param_prop)



import numpy as np
import birdepy.utility as ut
from birdepy.interface_probability import probability
from scipy.linalg import expm
import birdepy.utility_em as ut_em


def aug_bld_expm(sorted_data, param, b_rate, d_rate, likelihood, model, z_trunc,
                 j_tol, h_tol, options):
    """
    Creates a function based on matrix exponentials that returns the
    expected values of the number of up jumps from a state z, the number
    of down jumps from z, and the time spent in z when the process
    transitions from z0 to zt in elapsed time t (for any t that is a key
    in the dictionary `sorted_data`).

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.
    """

    z_min, z_max = z_trunc

    q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)

    num_states = int(z_max - z_min + 1)

    zero_mat = np.zeros((num_states, num_states))
    c_mat = np.vstack((np.hstack((q_mat, zero_mat)),
                       np.hstack((zero_mat, q_mat))))

    aug_data = {}
    for t in sorted_data:
        if likelihood in ['Erlang', 'expm', 'uniform']:
            p_mat = probability(np.arange(z_min, z_max + 1, 1),
                                np.arange(z_min, z_max + 1, 1), t, param,
                                method=likelihood, model=model,
                                z_trunc=z_trunc, options=options)
        for zz in sorted_data[t]:
            z0 = zz[0]
            zt = zz[1]
            if likelihood in ['Erlang', 'expm', 'uniform']:
                pr = p_mat[z0 - z_min, zt - z_min]
            elif likelihood in ['da', 'gwa', 'gwasa', 'ilt', 'oua']:
                pr = probability(z0, zt, t, param, model, likelihood,
                                 z_trunc=z_trunc, options=options)[0][0]
            c_mat[zt - z_min, num_states + z0 - z_min] = 1
            conv_mat = expm(c_mat*t)[0:num_states, num_states:]
            c_mat[zt - z_min, num_states + z0 - z_min] = 0  # Clean up for reuse

            def udh(z_):
                """
                Computes the expected values defined by Equation (14) in
                reference [1].

                [1] Crawford, F. W., Minin, V. N., & Suchard, M. A. (2014).
                Estimation for general birth-death processes. Journal of
                the American Statistical Association, 109(506), 730-747.

                """
                u = b_rate(z_, param) * \
                    conv_mat[z_ + 1 - z_min, z_ - z_min] / pr
                d = d_rate(z_, param) * \
                    conv_mat[z_ - 1 - z_min, z_ - z_min] / pr
                h = conv_mat[z_ - z_min, z_ - z_min] / pr
                return sorted_data[t][zz]*np.array([u, d, h])

            aug_data = ut_em.help_bld_aug_2(udh, z0, zt, aug_data, j_tol, h_tol,
                                            z_trunc)
    return aug_data



import numpy as np
import mpmath as mp
import birdepy.probability_expm as probability_expm
import birdepy.probability_uniform as probability_uniform
import birdepy.probability_Erlang as probability_Erlang
import birdepy.probability_gwa as probability_gwa
import birdepy.probability_gwasa as probability_gwasa
import birdepy.probability_oua as probability_oua
import birdepy.probability_da as probability_da
import birdepy.probability_ilt as probability_ilt
import birdepy.probability_sim as probability_sim
import birdepy.utility as ut


def probability(z0, zt, t, param, model='Verhulst 1', method='expm', **options):
    """Transition probabilities for continuous-time birth-and-death processes.

    Parameters
    ----------
    z0 : array_like
        States of birth-and-death process at time 0.

    zt : array_like
        States of birth-and-death process at time(s) `t`.

    t : array_like
        Elapsed time(s) (if has size greater than 1, then must be increasing).

    param : array_like
        The parameters governing the evolution of the birth-and-death
        process.
        Array of real elements of size (n,), where ‘n’ is the number of
        parameters.
        These must be in the order given :ref:`here
        <Birth-and-death Processes>`).

    model : string, optional
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst' (default)
            - 'Ricker'
            - 'Beverton-Holt'
            - 'Hassell'
            - 'MS-S'
            - 'Moran'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'MS-S'
            - 'M/M/inf'
            - 'loss-system'
            - 'custom'

        If set to 'custom', then arguments `b_rate` and `d_rate` must also be
        specified. See :ref:`here <Custom Models>` for more information.

    method : string, optional
        Transition probability approximation method. Should be one of
        (alphabetical ordering):

            - 'da' (see :ref:`here <birdepy.probability(method='da')>`)
            - 'Erlang'(see :ref:`here <birdepy.probability(method='Erlang')>`)
            - 'expm'  (default) (see :ref:`here <birdepy.probability(method='expm')>`)
            - 'gwa' (see :ref:`here <birdepy.probability(method='gwa')>`)
            - 'gwasa' (see :ref:`here <birdepy.probability(method='gwasa')>`)
            - 'ilt' (see :ref:`here <birdepy.probability(method='ilt')>`)
            - 'oua' (see :ref:`here <birdepy.probability(method='oua')>`)
            - 'sim' (see  :ref:`here <birdepy.probability(method='sim')>`)
            - 'uniform' (see :ref:`here <birdepy.probability(method='uniform')>`)

    options : dict, optional
        A dictionary of method specific options.

    Return
    -------
    transition_probability : ndarray
        An array of transition probabilities. If t has size bigger than 1,
        then the first coordinate corresponds to `t`, the second coordinate
        corresponds to `z0` and the third coordinate corresponds to `zt`;
        for example if `z0=[1,3,5,10]`, `zt=[5,8]` and `t=[1,2,3]`, then
        `transition_probability[2,0,1]` corresponds to
        P(Z(3)=8 | Z(0)=1).
        If `t` has size 1 the first coordinate corresponds to `z0` and the second
        coordinate corresponds to `zt`.

    Examples
    --------
    Approximate transition probability for a Verhulst model using various methods: ::

        for method in ['da', 'Erlang', 'expm', 'gwa', 'gwasa', 'ilt', 'oua', 'uniform']:
            print(method, 'approximation:',
            bd.probability(19, 27, 1.0, [0.5, 0.3, 0.01, 0.01], model='Verhulst', method=method)[0][0])

    Outputs: ::

        da approximation: 0.016040426614336103
        Erlang approximation: 0.0161337966847677
        expm approximation: 0.016191045449304713
        gwa approximation: 0.014646030484734228
        gwasa approximation: 0.014622270048744283
        ilt approximation: 0.01618465415009876
        oua approximation: 0.021627234315268227
        uniform approximation: 0.016191045442910168

    Notes
    -----
    Methods for computing transition probabilities and models are also
    described in [1]. If you use this function for published work, then please
    cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    :func:`birdepy.gpu_functions.probability()`  :func:`birdepy.gpu_functions.discrete()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    # Key word arguments can be passed as a dictionary which is itself the value
    # of a key word argument with key 'options', in this case the dictionary
    # must be extracted:
    if 'options' in options.keys():
        options = options['options']

    # Define the model-dependent birth and death rate functions
    if model == 'custom':
        b_rate = options['b_rate']
        d_rate = options['d_rate']
    else:
        b_rate = ut.higher_birth(model)
        d_rate = ut.higher_death(model)

    # Convert the arguments of 'z0', 'zt' and 't' into numpy arrays
    if np.isscalar(z0):
        z0 = np.array([z0])
    else:
        z0 = np.array(z0)
    if np.isscalar(zt):
        zt = np.array([zt])
    else:
        zt = np.array(zt)
    if np.isscalar(t):
        t = np.array([t])
    else:
        t = np.array(t)

    # Each of the methods for approximating transition probabilities are contained in their own
    # module probability_* where * depends on the method. This control structure determines which
    # module to use depending on the value of 'method'.  Method specific key word arguments are
    # extracted from the dictionary 'options' and then used in the module associated with the
    # method.
    if method == 'da':
        if 'k' in options.keys():
            k = options['k']
        else:
            k = 10
        if 'h_fun' in options.keys():
            h_fun = options['h_fun']
        else:
            h_fun = ut.higher_h_fun(model)
        return probability_da.probability_da(z0, zt, t, param, b_rate, d_rate,
                                             h_fun, k)

    elif method == 'Erlang':
        if 'z_trunc' in options.keys():
            z_trunc = options['z_trunc']
        else:
            z_vals = np.hstack((z0, zt))
            z_trunc = [np.maximum(0, np.amin(z_vals) - 100),
                       np.amax(z_vals) + 100]
        if 'k' in options.keys():
            k = options['k']
        else:
            k = 150
        return probability_Erlang.probability_Erlang(z0, zt, t, param, b_rate,
                                                     d_rate, z_trunc, k)

    elif method == 'expm':
        if 'z_trunc' in options.keys():
            z_trunc = options['z_trunc']
        else:
            z_vals = np.hstack((z0, zt))
            z_trunc = [np.maximum(0, np.amin(z_vals) - 100),
                       np.amax(z_vals) + 100]
        return probability_expm.probability_expm(z0, zt, t, param, b_rate,
                                                 d_rate,
                                                 z_trunc)

    elif method == 'gwa':
        if 'anchor' in options.keys():
            anchor = options['anchor']
        else:
            anchor = 'midpoint'
        return probability_gwa.probability_gwa(z0, zt, t, param, b_rate, d_rate, anchor)

    elif method == 'gwasa':
        if 'anchor' in options.keys():
            anchor = options['anchor']
        else:
            anchor = 'midpoint'
        return probability_gwasa.probability_gwasa(z0, zt, t, param, b_rate,
                                                   d_rate, anchor)

    elif method == 'ilt':
        if 'laplace_method' in options.keys():
            laplace_method = options['laplace_method']
        else:
            laplace_method = 'cme-talbot'
        if 'lentz_eps' in options.keys():
            lentz_eps = options['lentz_eps']
        else:
            lentz_eps = 1e-6
        if 'precision' in options.keys():
            mp.dps = options['precision']
        else:
            mp.dps = 1000
        if 'k' in options.keys():
            k = options['k']
        else:
            k = 25

        return probability_ilt.probability_ilt(z0, zt, t, param, b_rate,
                                               d_rate,
                                               lentz_eps, laplace_method, k)

    elif method == 'oua':
        if 'h_fun' in options.keys():
            h_fun = options['h_fun']
        else:
            h_fun = ut.higher_h_fun(model)
        if 'zf_bld' in options.keys():
            zf_bld = options['zf_bld']
        else:
            zf_bld = ut.higher_zf_bld(model)
        return probability_oua.probability_oua(z0, zt, t, param, b_rate,
                                               d_rate,
                                               h_fun, zf_bld)

    elif method == 'sim':
        if 'k' in options.keys():
            k = options['k']
        else:
            k = 10**6
        if 'seed' in options.keys():
            seed = options['seed']
        else:
            seed = None
        if 'sim_method' in options.keys():
            sim_method = options['sim_method']
        else:
            sim_method = 'exact'
        if 'tau' in options.keys():
            tau = options['tau']
        else:
            tau = 0.1
        return probability_sim.probability_sim(z0, zt, t, param, b_rate,
                                               d_rate, k, sim_method, tau, seed)

    elif method == 'uniform':
        if 'z_trunc' in options.keys():
            z_trunc = options['z_trunc']
        else:
            z_vals = np.hstack((z0, zt))
            z_trunc = [np.maximum(0, np.amin(z_vals) - 100),
                       np.amax(z_vals) + 100]
        if 'k' in options.keys():
            k = options['k']
        else:
            k = 150
        return probability_uniform.probability_uniform(z0, zt, t, param, b_rate,
                                                       d_rate, z_trunc, k)
    else:
        raise TypeError("Argument of 'method' is unknown, it should be one of "
                        "'da', 'Erlang', 'gwa', 'gwasa', 'ilt', 'expm', 'oua', "
                        "'sim', or 'uniform'. ")



import numpy as np
from birdepy.interface_probability import probability
import birdepy.utility_em as utem
import birdepy.utility as ut
import scipy.signal


def aug_bld_num(sorted_data, param, b_rate, d_rate, likelihood, model, z_trunc,
                j_tol, h_tol, options):
    """
    Creates a function based on numerical integration that returns the
    expected values of the number of up jumps from a state z, the number
    of down jumps from z, and the time spent in z when the process
    transitions from z0 to zt in elapsed time t (for any t that is a key
    in the dictionary `sorted_data`).

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.
    """
    if 'r' in options.keys():
        r = options['r']
    else:
        r = 10

    z_min, z_max = z_trunc

    states = np.arange(z_min, z_max+1, 1)

    p_mats = {}
    for _t in sorted_data:
        t_linspace = np.linspace(0, _t, 2 ** r + 1)
        output = np.zeros((t_linspace.size, states.size, states.size))
        p_mat_1 = probability(states, states, t_linspace[1], param, model,
                              likelihood, z_trunc=z_trunc, options=options)

        np.nan_to_num(p_mat_1, copy=False, nan=0)
        for row_idx in range(states.size):
            p_mat_1[row_idx, :] = scipy.signal.medfilt(p_mat_1[row_idx, :])
            row_sum = sum(p_mat_1[row_idx, :])
            if row_sum > 0:
                p_mat_1[row_idx, :] = p_mat_1[row_idx, :]/row_sum
            else:
                p_mat_1[row_idx, :] = 1/states.size
        output[0, :, :] = p_mat_1[np.ix_(np.array(states - z_min, dtype=np.int32),
                                         np.array(states - z_min, dtype=np.int32))]
        p_mat_power = p_mat_1
        for idx in range(1, t_linspace.size):
            p_mat_power = np.matmul(p_mat_power, p_mat_1)
            output[idx, :, :] = p_mat_power[np.ix_(np.array(states - z_min, dtype=np.int32),
                                            np.array(states - z_min, dtype=np.int32))]
        p_mats[_t] = output


    def udh(z0, zt, t, z):
        """
        Computes the expected values defined by Equation (14) in
        reference [1] using numerical integration.

        [1] Crawford, F. W., Minin, V. N., & Suchard, M. A. (2014).
        Estimation for general birth-death processes. Journal of
        the American Statistical Association, 109(506), 730-747.

        """
        t_linspace_ = np.linspace(0, t, 2 ** r + 1)
        pr = p_mats[t][-1, z0 - z_min, zt - z_min]
        probs1 = p_mats[t][:, z0 - z_min, z - z_min]
        probs2 = np.flip(p_mats[t][:, z - z_min + 1, zt - z_min])
        u = ut.trap_int(b_rate(z, param) * np.multiply(probs1, probs2) / pr,
                        t_linspace_)[-1]
        probs3 = np.flip(p_mats[t][:, z - z_min - 1, zt - z_min])
        d = ut.trap_int(d_rate(z, param) * np.multiply(probs1, probs3) / pr,
                        t_linspace_)[-1]
        probs4 = np.flip(p_mats[t][:, z - z_min, zt - z_min])
        h = ut.trap_int(np.multiply(probs1, probs4) / pr,
                        t_linspace_)[-1]
        return np.array([u, d, h])

    aug_data = utem.help_bld_aug(udh, sorted_data, j_tol, h_tol, z_trunc)

    return aug_data
